"""
Most classes in this directory are imported from the main module,
but there are some other utility classes here
"""

from .entity.data_file import DataFile
from .entity.download_data_file import (
    DownloadDataFile,
    DownloadDataFileToDisk,
    DownloadDataFileToMemory,
    DownloadDataFileToMemoryAndDisk,
)
from .entity.data_file_directory import DataFileDirectory
from .entity.reproducer_message import ReproducerMessage

__all__ = [
    "DataFile",
    "DownloadDataFile",
    "DownloadDataFileToDisk",
    "DownloadDataFileToMemory",
    "DownloadDataFileToMemoryAndDisk",
    "DataFileDirectory",
    "ReproducerMessage",
]



"""Script to invoke the OS-appropriate ServiceManager class to do anything other than installing"""

# imports
from .config import SERVICE_CONST
from .utilities import get_os_name
from .service_manager_base import ServiceManagerBase
from .windows_service_manager import WindowsServiceManager
from .linux_service_manager import LinuxServiceManager


def main(given_args=None):
    """
    Use the OS-appropriate ServiceManager class to manage a Service/daemon
    (i.e., anything except installing it)
    """
    # get the arguments
    parser = ServiceManagerBase.get_argument_parser("manage")
    args = parser.parse_args() if given_args is None else parser.parse_args(given_args)
    # get the name of the OS and start the object
    operating_system = get_os_name()
    manager_args = [args.service_name]
    manager_kwargs = {"interactive": given_args is None, "logger": SERVICE_CONST.logger}
    managers_by_os_name = {
        "Windows": WindowsServiceManager,
        "Linux": LinuxServiceManager,
    }
    manager_class = managers_by_os_name[operating_system]
    manager = manager_class(*manager_args, **manager_kwargs)
    # run some function based on the run mode
    manager.run_manage_command(
        args.run_mode,
        remove_env_vars=args.remove_env_vars,
        remove_install_args=args.remove_install_args,
        remove_nssm=args.remove_nssm,
    )


if __name__ == "__main__":
    main()



"""A set of Producers sharing an optional KafkaCrypto instance for key passing"""

# imports
from openmsitoolbox import LogOwner
from .openmsistream_producer import OpenMSIStreamProducer


class ProducerGroup(LogOwner):
    """
    Class for working with a group of producers sharing a single
    :class:`kafkacrypto.KafkaCrypto` instance

    :param config_path: Path to the config file that should be used to define Producers in the group
    :type config_path: :class:`pathlib.Path`
    :param kafkacrypto: The :class:`~OpenMSIStreamKafkaCrypto` object that should be used
        to instantiate Producers. Only needed if a single specific
        :class:`~OpenMSIStreamKafkaCrypto` instance should be shared.
    :type kafkacrypto: :class:`~OpenMSIStreamKafkaCrypto`, optional
    """

    @property
    def kafkacrypto(self):
        """
        The KafkaCrypto object handling key passing and serialization for encrypted messages
        """
        return (
            self.__p_kwargs["kafkacrypto"] if "kafkacrypto" in self.__p_kwargs else None
        )

    def __init__(self, config_path, kafkacrypto=None, **kwargs):
        """
        Constructor method
        """
        super().__init__(**kwargs)
        self.__p_args, self.__p_kwargs = OpenMSIStreamProducer.get_producer_args_kwargs(
            config_path, logger=self.logger, kafkacrypto=kafkacrypto
        )

    def get_new_producer(self):
        """
        Return a new :class:`~OpenMSIStreamProducer` object.
        Call this function from a child thread to get thread-independent Producers.
        Note: this function just creates the Producer; closing it etc. must be handled
        by whatever calls this function.

        :return: a Producer created using the config set in the constructor
        :rtype: :class:`~OpenMSIStreamProducer`
        """
        producer = OpenMSIStreamProducer(*self.__p_args, **self.__p_kwargs)
        return producer

    def close(self):
        """
        Wrapper around :func:`kafkacrypto.KafkaCrypto.close`.
        """
        try:
            self.__p_kwargs["kafkacrypto"].close()
        except Exception:
            pass
        finally:
            self.__p_kwargs["kafkacrypto"] = None

    @classmethod
    def get_command_line_arguments(cls):
        """
        Anything extending this class should be able to access the
        "treat_undecryptable_as_plaintext" flag
        """
        superargs, superkwargs = super().get_command_line_arguments()
        args = [
            *superargs,
            "config",
        ]
        return args, superkwargs

    @classmethod
    def get_init_args_kwargs(cls, parsed_args):
        superargs, superkwargs = super().get_init_args_kwargs(parsed_args)
        args = [
            parsed_args.config,
            *superargs,
        ]
        return args, superkwargs



"""
Code for managing the two .csv files listing files that are
being reconstructed from the topic and files that have been fully handled
"""

# imports
import pathlib, datetime, re, threading
from abc import ABC
from dataclasses import dataclass
from openmsitoolbox import LogOwner
from ....utilities import (
    DataclassTableReadOnly,
    DataclassTableAppendOnly,
    DataclassTable,
)
from ...utilities import get_message_prepend


@dataclass
class StreamHandlerRegistryLineInProgress:
    """
    A line in the table listing files in progress
    """

    filename: str
    rel_filepath: pathlib.Path
    status: str
    n_chunks: int
    first_message: datetime.datetime
    most_recent_message: datetime.datetime


@dataclass
class StreamHandlerRegistryLineSucceeded:
    """
    A line in the table listing files that have been successfully handled
    """

    filename: str
    rel_filepath: pathlib.Path
    n_chunks: int
    first_message: datetime.datetime
    processed_at: datetime.datetime


class StreamHandlerRegistry(LogOwner, ABC):
    """
    Keeps track of the status of files read during stream handling
    """

    IN_PROGRESS = "in_progress"
    MISMATCHED_HASH = "mismatched_hash"

    @property
    def in_progress_table(self):
        """
        A read-only version of the table listing files in progress
        """
        return self._in_progress_table.as_read_only()

    @property
    def succeeded_table(self):
        """
        A read-only version of the table listing files that have been successfully handled
        """
        self.consolidate_succeeded_files()
        return DataclassTableReadOnly(
            StreamHandlerRegistryLineSucceeded,
            filepath=self.succeeded_filepath,
            logger=self.logger,
        )

    @property
    def filepaths_to_rerun(self):
        """
        A list of all the filepaths that have not yet been successfully processed
        """
        to_rerun = []
        for addr in self._in_progress_table.obj_addresses:
            to_rerun.append(self._in_progress_table.get_entry_attrs(addr, "rel_filepath"))
        return to_rerun

    @property
    def rerun_file_key_regex(self):
        """
        Regex matching keys for messages from files that need to be rerun
        """
        if self.n_files_to_rerun <= 0:
            return None
        regex_str = r"^("
        for fp in self.filepaths_to_rerun:
            subdir_str = str(fp.parent.as_posix()) if fp != fp.name else None
            msg_prepend = get_message_prepend(subdir_str, fp.name)
            regex_part = re.escape(msg_prepend)
            regex_str += f"{regex_part}|"
        regex_str = f"{regex_str[:-1]})_.*$"
        return re.compile(regex_str)

    @property
    def n_files_to_rerun(self):
        """
        The number of files in the registry that are marked as "in_progress"
        """
        return len(self.filepaths_to_rerun)

    def __init__(
        self,
        in_progress_filepath,
        succeeded_filepath,
        *args,
        max_succeeded_lines_per_file=1000,
        **kwargs,
    ):
        """
        in_progress_filepath = path to the file with the "in_progress" datatable
        succeeded_filepath = path to the file with the "succeeded" datatable
        """
        super().__init__(*args, **kwargs)
        self._in_progress_table = DataclassTable(
            dataclass_type=StreamHandlerRegistryLineInProgress,
            filepath=in_progress_filepath,
            logger=self.logger,
        )
        self.succeeded_filepath = succeeded_filepath
        self.max_succeeded_lines_per_file = max_succeeded_lines_per_file
        self.__succeeded_tables_by_id = {}
        self.consolidate_succeeded_files()

    def register_file_in_progress(self, dfc):
        """
        Add/update a line in the table showing that a file is in progress
        """
        self._add_or_modify_in_progress_entry(dfc, self.IN_PROGRESS)

    def register_file_mismatched_hash(self, dfc):
        """
        Add/update a line in the table to show that a consumed file
        was mismatched with its original content hash
        """
        self._add_or_modify_in_progress_entry(dfc, self.MISMATCHED_HASH)

    def consolidate_succeeded_files(self):
        """
        Search the directory holding the files for all "completed" file entries
        and write them into a single file
        """
        # add any lines from files in the directory to the one consolidated file
        globpattern = f"{self.succeeded_filepath.stem}*{self.succeeded_filepath.suffix}"
        consolidated_file = None
        for fp in self.succeeded_filepath.parent.glob(globpattern):
            if fp == self.succeeded_filepath:
                continue
            for table in self.__succeeded_tables_by_id.values():
                if fp == table.filepath:
                    table.dump_to_file()
            file_to_add = DataclassTableReadOnly(
                dataclass_type=StreamHandlerRegistryLineSucceeded,
                filepath=fp,
                logger=self.logger,
            )
            if consolidated_file is None:
                consolidated_file = DataclassTableAppendOnly(
                    dataclass_type=StreamHandlerRegistryLineSucceeded,
                    filepath=self.succeeded_filepath,
                    logger=self.logger,
                )
            consolidated_file.add_entries(file_to_add.objects)
        if consolidated_file is None:
            return
        # dump out the consolidated file
        consolidated_file.dump_to_file()
        # make sure that all the lines from the individual files have been copied
        all_lines = (
            DataclassTableReadOnly(
                dataclass_type=StreamHandlerRegistryLineSucceeded,
                filepath=self.succeeded_filepath,
                logger=self.logger,
            )
        ).lines
        for fp in self.succeeded_filepath.parent.glob(globpattern):
            if fp == self.succeeded_filepath:
                continue
            added_file = DataclassTableReadOnly(
                dataclass_type=StreamHandlerRegistryLineSucceeded,
                filepath=fp,
                logger=self.logger,
            )
            for entry_line in added_file.lines:
                if entry_line not in all_lines:
                    warnmsg = (
                        "WARNING: failed to consolidate individual files into "
                        f"{self.succeeded_filepath}. Individual files will be retained "
                        "and should be manually concatenated. "
                        "Duplicate entries may be present."
                    )
                    self.logger.warning(warnmsg)
            to_pop = [
                table_id
                for table_id, table in self.__succeeded_tables_by_id.items()
                if fp == table.filepath
            ]
            for table_id in to_pop:
                self.__succeeded_tables_by_id.pop(table_id)
            fp.unlink()

    def _add_or_modify_in_progress_entry(self, dfc, new_status):
        with self._in_progress_table.lock:
            existing_entry_addr = self._get_in_progress_address_for_rel_filepath(
                dfc.relative_filepath
            )
            if existing_entry_addr is None:
                new_entry = StreamHandlerRegistryLineInProgress(
                    dfc.filename,
                    dfc.relative_filepath,
                    new_status,
                    dfc.n_total_chunks,
                    datetime.datetime.now(),
                    datetime.datetime.now(),
                )
                self._in_progress_table.add_entries(new_entry)
                self._in_progress_table.dump_to_file()
            else:
                self._in_progress_table.set_entry_attrs(
                    existing_entry_addr,
                    status=new_status,
                    most_recent_message=datetime.datetime.now(),
                )

    def _get_in_progress_address_for_rel_filepath(self, rel_filepath):
        existing_obj_addresses = self._in_progress_table.obj_addresses_by_key_attr(
            "rel_filepath"
        )
        if rel_filepath not in existing_obj_addresses:
            return None
        if len(existing_obj_addresses[rel_filepath]) != 1:
            errmsg = (
                f"ERROR: found more than one {self.__class__.__name__} entry for "
                f"relative filepath {rel_filepath} in file "
                f"at {self._in_progress_table.filepath}"
            )
            self.logger.error(errmsg, exc_type=ValueError)
        return existing_obj_addresses[rel_filepath][0]

    def _add_to_succeeded_table(self, new_entry, thread_identifier=None):
        if thread_identifier is None:
            thread_identifier = f"t{threading.get_ident()}"
        if thread_identifier in self.__succeeded_tables_by_id:
            if (
                self.__succeeded_tables_by_id[thread_identifier].n_entries
                >= self.max_succeeded_lines_per_file
            ):
                old_fp = self.__succeeded_tables_by_id[thread_identifier].filepath
                timestamp = str(datetime.datetime.now().timestamp()).replace(".", "_")
                new_fp = old_fp.with_name(f"{old_fp.stem}_{timestamp}{old_fp.suffix}")
                old_fp.rename(new_fp)
                new_table = DataclassTableAppendOnly(
                    StreamHandlerRegistryLineSucceeded,
                    filepath=old_fp,
                    logger=self.logger,
                )
                self.__succeeded_tables_by_id[thread_identifier] = new_table
        else:
            stem = self.succeeded_filepath.stem
            suffix = self.succeeded_filepath.suffix
            new_name = f"{stem}_{thread_identifier}{suffix}"
            new_fp = self.succeeded_filepath.with_name(new_name)
            new_table = DataclassTableAppendOnly(
                StreamHandlerRegistryLineSucceeded, filepath=new_fp, logger=self.logger
            )
            self.__succeeded_tables_by_id[thread_identifier] = new_table
        self.__succeeded_tables_by_id[thread_identifier].add_entries(new_entry)
        self.__succeeded_tables_by_id[thread_identifier].dump_to_file()


class StreamProcessorRegistry(StreamHandlerRegistry):
    """
    Keeps track of the status of files read during stream processing
    """

    FAILED = "failed"

    def __init__(self, dirpath, topic_name, consumer_group_id, *args, **kwargs):
        """
        dirpath    = path to the directory that should contain the csv files
        topic_name = the name of the topic that will be produced to
        """
        in_progress_filepath = (
            dirpath
            / f"consuming_from_{topic_name}_in_progress_by_{consumer_group_id}.csv"
        )
        succeeded_filepath = (
            dirpath / f"processed_from_{topic_name}_by_{consumer_group_id}.csv"
        )
        super().__init__(in_progress_filepath, succeeded_filepath, *args, **kwargs)

    def register_file_successfully_processed(self, dfc):
        """
        Add/update a line in the table to show that a file has successfully been processed
        """
        self._in_progress_table.lock.acquire()
        existing_entry_addr = self._get_in_progress_address_for_rel_filepath(
            dfc.relative_filepath
        )
        if existing_entry_addr is not None:
            attrs = self._in_progress_table.get_entry_attrs(existing_entry_addr)
            new_entry = StreamHandlerRegistryLineSucceeded(
                attrs["filename"],
                attrs["rel_filepath"],
                attrs["n_chunks"],
                attrs["first_message"],
                datetime.datetime.now(),
            )
            self._add_to_succeeded_table(new_entry)
            try:
                self._in_progress_table.remove_entries(existing_entry_addr)
                self._in_progress_table.dump_to_file()
            except ValueError:
                pass
        else:
            new_entry = StreamHandlerRegistryLineSucceeded(
                dfc.filename,
                dfc.relative_filepath,
                dfc.n_total_chunks,
                datetime.datetime.now(),
                datetime.datetime.now(),
            )
            self._add_to_succeeded_table(new_entry)
        self._in_progress_table.lock.release()

    def register_file_processing_failed(self, dfc):
        """
        Add/update a line in the table to show that a file has failed to be processed
        """
        self._add_or_modify_in_progress_entry(dfc, self.FAILED)


class StreamReproducerRegistry(StreamHandlerRegistry):
    """
    Keeps track of the status of files read from one topic with associated
    information created and re-produced to a different topic
    """

    PRODUCING_MESSAGE_FAILED = "producing_message_failed"
    COMPUTING_RESULT_FAILED = "computing_result_message_failed"

    def __init__(
        self,
        dirpath,
        consumer_topic_name,
        producer_topic_name,
        consumer_group_id,
        *args,
        **kwargs,
    ):
        """
        dirpath    = path to the directory that should contain the csv file
        topic_name = the name of the topic that will be produced to (used in the filename)
        """
        in_progress_filepath = (
            dirpath
            / f"consuming_from_{consumer_topic_name}_in_progress_by_{consumer_group_id}.csv"
        )
        succeeded_filepath = dirpath / f"results_produced_to_{producer_topic_name}.csv"
        super().__init__(in_progress_filepath, succeeded_filepath, *args, **kwargs)

    def register_file_results_produced(
        self, filename, rel_filepath, n_total_chunks, prodid
    ):
        """
        Add/update a line in the table to show that a file has successfully been processed
        """
        self._in_progress_table.lock.acquire()
        existing_entry_addr = self._get_in_progress_address_for_rel_filepath(rel_filepath)
        if existing_entry_addr is not None:
            attrs = self._in_progress_table.get_entry_attrs(existing_entry_addr)
            new_entry = StreamHandlerRegistryLineSucceeded(
                attrs["filename"],
                attrs["rel_filepath"],
                attrs["n_chunks"],
                attrs["first_message"],
                datetime.datetime.now(),
            )
            self._add_to_succeeded_table(new_entry, f"p{prodid}")
            try:
                self._in_progress_table.remove_entries(existing_entry_addr)
                self._in_progress_table.dump_to_file()
            except ValueError:
                pass
        else:
            new_entry = StreamHandlerRegistryLineSucceeded(
                filename,
                rel_filepath,
                n_total_chunks,
                datetime.datetime.now(),
                datetime.datetime.now(),
            )
            self._add_to_succeeded_table(new_entry, f"p{prodid}")
        self._in_progress_table.lock.release()

    def register_file_computing_result_failed(
        self, filename, rel_filepath, n_total_chunks
    ):
        """
        Update a line in the table to show that a file failed to have its processing result computed
        """
        self._add_or_modify_in_progress_entry_without_chunk(
            filename, rel_filepath, n_total_chunks, self.COMPUTING_RESULT_FAILED
        )

    def register_file_result_production_failed(
        self, filename, rel_filepath, n_total_chunks
    ):
        """
        Update a line in the table to show that the call to Producer.produce()
         for a message computed from a file returned an error
        """
        self._add_or_modify_in_progress_entry_without_chunk(
            filename, rel_filepath, n_total_chunks, self.PRODUCING_MESSAGE_FAILED
        )

    def _add_or_modify_in_progress_entry_without_chunk(
        self, filename, rel_filepath, n_total_chunks, new_status
    ):
        with self._in_progress_table.lock:
            existing_entry_addr = self._get_in_progress_address_for_rel_filepath(
                rel_filepath
            )
            if existing_entry_addr is None:
                new_entry = StreamHandlerRegistryLineInProgress(
                    filename,
                    rel_filepath,
                    new_status,
                    n_total_chunks,
                    datetime.datetime.now(),
                    datetime.datetime.now(),
                )
                self._in_progress_table.add_entries(new_entry)
                self._in_progress_table.dump_to_file()
            else:
                self._in_progress_table.set_entry_attrs(
                    existing_entry_addr,
                    status=new_status,
                    most_recent_message=datetime.datetime.now(),
                )



" Various Producibles for main program/base class heartbeat messages "

# imports
import datetime, json
from ..kafka_wrapper import Producible


class HeartbeatProducible(Producible):
    """A Producible for heartbeat messages. The key is a program ID and the value
    is a timestamp
    """

    def __init__(self, program_id):
        self.__program_id = program_id

    @property
    def msg_key(self):
        return f"{self.__program_id}_heartbeat"

    @property
    def msg_value(self):
        value_dict = {"timestamp": str(datetime.datetime.now())}
        return json.dumps(value_dict)


class UploadDirectoryHeartbeatProducible(HeartbeatProducible):
    "The heartbeat Producible sent by DataFileUploadDirectory programs."

    def __init__(self, program_id, n_msgs, n_bytes):
        super().__init__(program_id)
        self.__n_msgs = n_msgs
        self.__n_bytes = n_bytes

    @property
    def msg_value(self):
        value_dict = json.loads(super().msg_value)
        value_dict["n_messages_produced"] = self.__n_msgs
        value_dict["n_bytes_produced"] = self.__n_bytes
        return json.dumps(value_dict)


class MessageProcessorHeartbeatProducible(HeartbeatProducible):
    """The heartbeat Producible sent by any programs that extend the
    ControlledMessageProcessor class
    """

    def __init__(
        self, program_id, n_msgs_read, n_msgs_processed, n_bytes_read, n_bytes_processed
    ):
        super().__init__(program_id)
        self.__n_msgs_read = n_msgs_read
        self.__n_msgs_processed = n_msgs_processed
        self.__n_bytes_read = n_bytes_read
        self.__n_bytes_processed = n_bytes_processed

    @property
    def msg_value(self):
        value_dict = json.loads(super().msg_value)
        value_dict["n_messages_read"] = self.__n_msgs_read
        value_dict["n_messages_processed"] = self.__n_msgs_processed
        value_dict["n_bytes_read"] = self.__n_bytes_read
        value_dict["n_bytes_processed"] = self.__n_bytes_processed
        return json.dumps(value_dict)


class MessageReproducerHeartbeatProducible(HeartbeatProducible):
    """The heartbeat Producible sent by any programs that extend the
    ControlledMessageReproducer class
    """

    def __init__(
        self,
        program_id,
        n_msgs_read,
        n_msgs_processed,
        n_msgs_produced,
        n_bytes_read,
        n_bytes_processed,
        n_bytes_produced,
    ):
        super().__init__(program_id)
        self.__n_msgs_read = n_msgs_read
        self.__n_msgs_processed = n_msgs_processed
        self.__n_msgs_produced = n_msgs_produced
        self.__n_bytes_read = n_bytes_read
        self.__n_bytes_processed = n_bytes_processed
        self.__n_bytes_produced = n_bytes_produced

    @property
    def msg_value(self):
        value_dict = json.loads(super().msg_value)
        value_dict["n_messages_read"] = self.__n_msgs_read
        value_dict["n_messages_processed"] = self.__n_msgs_processed
        value_dict["n_messages_produced"] = self.__n_msgs_produced
        value_dict["n_bytes_read"] = self.__n_bytes_read
        value_dict["n_bytes_processed"] = self.__n_bytes_processed
        value_dict["n_bytes_produced"] = self.__n_bytes_produced
        return json.dumps(value_dict)



"""Classes for serialization and deserialization operations"""

# imports
import pathlib, inspect, time, warnings
from hashlib import sha512
import msgpack
from confluent_kafka.error import SerializationError
from confluent_kafka.serialization import Serializer, Deserializer

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from kafkacrypto.message import KafkaCryptoMessage, KafkaCryptoMessageError
from ..data_file_io.entity.data_file_chunk import DataFileChunk

####################### COMPOUND (DE)SERIALIZERS FOR STACKING MULTIPLE STEPS #######################


class CompoundSerDes(Serializer, Deserializer):
    """
    A Serializer/Deserializer that stacks multiple Serialization/Deserialization steps
    """

    def __init__(self, *args):
        """
        args = a list of (De)Serializer (or otherwise callable) objects to apply
        IN GIVEN ORDER to some data
        """
        self.__steps = list(args)

    def __call__(self, data, ctx=None):
        if data is None:
            return None
        for istep, serdes in enumerate(self.__steps, start=1):
            try:
                data = serdes(data)
            except Exception as exc:
                errmsg = (
                    f"ERROR: failed to (de)serialize at step {istep} (of {len(self.__steps)}) "
                    f"in CompoundSerDes! Callable = {serdes}, exception = {exc}"
                )
                raise SerializationError(errmsg)
        return data


class CompoundSerializer(Serializer):
    """
    A Serializer that stacks multiple steps
    For use with KafkaCrypto since topic names must be passed
    """

    def __init__(self, *args):
        self.__steps = list(args)

    def serialize(self, topic, data):
        """
        Call a chain of Serializers for a given message (may contain encryption)
        """
        for istep, ser in enumerate(self.__steps, start=1):
            try:
                if hasattr(ser, "serialize") and inspect.isroutine(ser.serialize):
                    data = ser.serialize(topic, data)
                else:
                    data = ser(data, ctx=None)
            except Exception as exc:
                errmsg = (
                    f"ERROR: failed to serialize at step {istep} (of {len(self.__steps)}) "
                    f"in {self.__class__.__name__}! Callable = {ser}, exception = {exc}"
                )
                raise SerializationError(errmsg)
        return data

    def __call__(self, data, ctx=None):
        if data is None:
            return None
        raise SerializationError(
            f"ERROR: __call__ method not implemented for a {self.__class__.__name__}"
        )


class CompoundDeserializer(Deserializer):
    """
    A Deserializer that stacks multiple steps
    For use with KafkaCrypto since topic names must be passed
    """

    MAX_WAIT_PER_DECRYPT = 60.0  # in seconds

    def __init__(self, *args, treat_undecryptable_as_plaintext=False):
        self.__steps = list(args)
        self.treat_undecryptable_as_plaintext = treat_undecryptable_as_plaintext

    def deserialize(self, topic, data):
        """
        Call a chain of Deserializers for a given message (may contain decryption)
        """
        for istep, des in enumerate(self.__steps, start=1):
            try:
                data = self.__deserialize_message(des, topic, data)
            except Exception as exc:
                errmsg = (
                    f"ERROR: failed to deserialize at step {istep} (of {len(self.__steps)}) "
                    f"in {self.__class__.__name__}! Callable = {des}, exception = {exc}"
                )
                raise SerializationError(errmsg)
        return data

    def __call__(self, data, ctx=None):
        if data is None:
            return None
        raise SerializationError(
            f"ERROR: __call__ method not implemented for a {self.__class__.__name__}"
        )

    def __deserialize_message(self, des, topic, data):
        """
        Deserialize a (possibly encrypted) message
        """
        # deserialize a KafkaCrypto message
        if hasattr(des, "deserialize") and inspect.isroutine(des.deserialize):
            data = des.deserialize(topic, data)
            if isinstance(data, KafkaCryptoMessage):
                success = False
                elapsed = 0
                while (
                    (not success)
                    and elapsed < self.MAX_WAIT_PER_DECRYPT
                    and data.isPossiblyDecryptable()
                ):
                    try:
                        data = data.getMessage()
                        success = True
                    except KafkaCryptoMessageError:
                        time.sleep(0.1)
                        elapsed += 0.1
                # if the message could not be decrypted, return the Message object
                # itself, or the raw bytes if never decryptable and tunable above
                # is True
                if not success:
                    if (
                        not data.isPossiblyDecryptable()
                    ) and self.treat_undecryptable_as_plaintext:
                        return bytes(data)
                    return data
        # if a previous step failed to deserialize a KafkaCrypto message,
        # don't bother trying this next step
        elif not isinstance(data, KafkaCryptoMessage):
            data = des(data, ctx=None)
        return data


####################### SERIALIZING/DESERIALIZING FILE CHUNKS #######################


class DataFileChunkSerializer(Serializer):
    """
    Serializes DataFileChunk objects to bytesarrays
    """

    def __call__(self, file_chunk_obj, ctx=None):
        if file_chunk_obj is None:
            return None
        if not isinstance(file_chunk_obj, DataFileChunk):
            raise SerializationError(
                "ERROR: object passed to FileChunkSerializer is not a DataFileChunk!"
            )
        # pack up all the relevant bits of information into a single bytearray
        try:
            # get the chunk's data from the file if necessary
            if file_chunk_obj.data is None:
                file_chunk_obj.populate_with_file_data()
            # pack up all of the properties of the message
            ordered_properties = []
            ordered_properties.append(str(file_chunk_obj.filename))
            ordered_properties.append(file_chunk_obj.file_hash)
            ordered_properties.append(file_chunk_obj.chunk_hash)
            ordered_properties.append(file_chunk_obj.chunk_offset_write)
            ordered_properties.append(file_chunk_obj.chunk_i)
            ordered_properties.append(file_chunk_obj.n_total_chunks)
            ordered_properties.append(file_chunk_obj.subdir_str)
            ordered_properties.append(file_chunk_obj.filename_append)
            ordered_properties.append(file_chunk_obj.data)
            return msgpack.packb(ordered_properties, use_bin_type=True)
        except Exception as exc:
            raise SerializationError(
                f"ERROR: failed to serialize a DataFileChunk! Exception: {exc}"
            )


class DataFileChunkDeserializer(Deserializer):
    """
    Deseralizes bytearrays to DataFileChunk objects
    """

    def __call__(self, byte_array, ctx=None):
        if byte_array is None:
            return None
        try:
            # unpack the byte array
            ordered_properties = msgpack.unpackb(byte_array, raw=True)
            if len(ordered_properties) != 9:
                errmsg = (
                    f"ERROR: unrecognized token passed to DataFileChunkDeserializer. "
                    f"Expected 9 properties but found {len(ordered_properties)}"
                )
                raise ValueError(errmsg)
            try:
                filename = str(ordered_properties[0].decode())
                file_hash = ordered_properties[1]
                chunk_hash = ordered_properties[2]
                chunk_offset_read = None
                chunk_offset_write = int(ordered_properties[3])
                chunk_i = int(ordered_properties[4])
                n_total_chunks = int(ordered_properties[5])
                subdir_str = str(ordered_properties[6].decode())
                filename_append = str(ordered_properties[7].decode())
                data = ordered_properties[8]
            except Exception as exc:
                errmsg = (
                    f"ERROR: unrecognized value(s) when deserializing a DataFileChunk "
                    f"from token. Exception: {exc}"
                )
                raise ValueError(errmsg)
            # make sure the hash of the chunk's data matches with what it was before
            check_chunk_hash = sha512()
            check_chunk_hash.update(data)
            check_chunk_hash = check_chunk_hash.digest()
            if check_chunk_hash != chunk_hash:
                errmsg = (
                    f"ERROR: chunk hash {check_chunk_hash} != expected hash {chunk_hash} "
                    f"in file {filename}, offset {chunk_offset_write}"
                )
                raise RuntimeError(errmsg)
            # set the filepath based on the subdirectory string
            if subdir_str == "":
                filepath = pathlib.Path(filename)
            subdir_path = pathlib.PurePosixPath(subdir_str)
            filepath = pathlib.Path("").joinpath(*(subdir_path.parts), filename)
            return DataFileChunk(
                filepath,
                filename,
                file_hash,
                chunk_hash,
                chunk_offset_read,
                chunk_offset_write,
                len(data),
                chunk_i,
                n_total_chunks,
                data=data,
                filename_append=filename_append,
            )
        except Exception as exc:
            raise SerializationError(
                f"ERROR: failed to deserialize a DataFileChunk! Exception: {exc}"
            )



"""A directory of DataFiles"""

# imports
from openmsitoolbox import LogOwner
from openmsitoolbox.utilities.misc import populated_kwargs


class DataFileDirectory(LogOwner):
    """
    Base class representing any directory holding data files

    :param dirpath: path to the directory
    :type dirpath: :class:`pathlib.Path`
    """

    def __init__(self, dirpath, *args, **kwargs):
        self.dirpath = dirpath.resolve()
        kwargs = populated_kwargs(kwargs, {"logger_file": self.dirpath})
        super().__init__(*args, **kwargs)



"""
A ConsumerGroup whose receipt of messages is governed using the ControlledProcess infrastructure
"""

# imports
import warnings
from abc import ABC, abstractmethod

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from kafkacrypto import KafkaCryptoMessage
    from kafkacrypto.confluent_kafka_wrapper import Message
from ..utilities.heartbeat_producibles import MessageProcessorHeartbeatProducible
from ..utilities.controlled_processes_heartbeats import (
    ControlledProcessMultiThreadedHeartbeats,
)
from .consumer_group import ConsumerGroup


class ControlledMessageProcessor(
    ControlledProcessMultiThreadedHeartbeats, ConsumerGroup, ABC
):
    """
    Combine a ControlledProcessMultiThreaded and a ConsumerGroup to create a
    single interface for reading and processing individual messages
    """

    CONSUMER_POLL_TIMEOUT = 5.0

    def __init__(self, config_path, topic_name, filepath_regex=None, **kwargs):
        """
        Hang onto the number of messages read and processed
        """
        self.n_msgs_read = 0
        self.n_msgs_processed = 0
        super().__init__(config_path, topic_name, **kwargs)
        # set below to true to reset new consumers to their earliest offsets
        self.restart_at_beginning = False
        # set below to some regex to filter messages by their keys
        self.message_key_regex = None
        # reset the key regex after the consumer has filtered through previous messages
        self.filter_new_message_keys = False
        # set below to some regex to filter messages by filepath
        self.filepath_regex = filepath_regex
        # hold onto the last consumed message to manually commit its offset on shutdown
        self.last_message = None
        # variables for heartbeat messages
        self.n_msgs_read_since_last_heartbeat = 0
        self.n_msgs_processed_since_last_heartbeat = 0
        self.n_bytes_read_since_last_heartbeat = 0
        self.n_bytes_processed_since_last_heartbeat = 0

    def get_heartbeat_message(self):
        new_msg = MessageProcessorHeartbeatProducible(
            self._heartbeat_program_id,
            self.n_msgs_read_since_last_heartbeat,
            self.n_msgs_processed_since_last_heartbeat,
            self.n_bytes_read_since_last_heartbeat,
            self.n_bytes_processed_since_last_heartbeat,
        )
        with self.lock:
            self.n_msgs_read_since_last_heartbeat = 0
            self.n_msgs_processed_since_last_heartbeat = 0
            self.n_bytes_read_since_last_heartbeat = 0
            self.n_bytes_processed_since_last_heartbeat = 0
        return new_msg

    def _run_worker(self):
        """
        Handle startup and shutdown of a thread-independent Consumer and
        serve individual messages to the _process_message function
        """
        # create the Consumer for this thread
        if self.alive:
            consumer = self.get_new_subscribed_consumer(
                restart_at_beginning=self.restart_at_beginning,
                message_key_regex=self.message_key_regex,
                filter_new_message_keys=self.filter_new_message_keys,
                filepath_regex=self.filepath_regex,
            )
        if ("enable.auto.commit" not in consumer.configs.keys()) or (
            consumer.configs["enable.auto.commit"] is True
        ):
            warnmsg = (
                "WARNING: enable.auto.commit has not been set to False for a Consumer "
                "that will manually commit offsets. Missed or duplicate messages could result. "
                'You can set "enable.auto.commit"=False in the "consumer" section of the '
                "config file to re-enable manual offset commits (recommended)."
            )
            self.logger.warning(warnmsg)
        # start the loop for while the controlled process is alive
        while self.alive:
            self.__do_alive_loop_iteration(consumer)
        # commit the offset of the last message received
        if (self.last_message is not None) and (
            not consumer.message_consumed_before(self.last_message)
        ):
            self.__commit_last_message_offset(consumer)
        # shut down the Consumer that was created once the process isn't alive anymore
        consumer.close()

    def _on_shutdown(self):
        self.close()
        super()._on_shutdown()

    @abstractmethod
    def _process_message(self, lock, msg, *args, **kwargs):
        """
        Process a single message read from the thread-independent Consumer
        Returns true if processing was successful, and False otherwise

        lock = lock across all created child threads
        msg  = a single message that was consumed and should be processed by this function

        Not implemented in the base class
        """
        raise NotImplementedError

    def __do_alive_loop_iteration(self, consumer):
        """
        Helper function to perform actions that happen continuously while the process is alive
        """
        # consume a message from the topic
        msg = consumer.get_next_message(self.CONSUMER_POLL_TIMEOUT)
        if msg is None:
            return
        with self.lock:
            self.n_msgs_read += 1
            self.n_msgs_read_since_last_heartbeat += 1
            if (
                hasattr(msg, "key")
                and hasattr(msg, "value")
                and (
                    isinstance(msg.key, KafkaCryptoMessage)
                    or isinstance(msg.value, KafkaCryptoMessage)
                )
            ):
                self.n_bytes_read_since_last_heartbeat += len(bytes(msg))
            elif isinstance(msg, Message):
                self.n_bytes_read_since_last_heartbeat += len(msg.value)
            else:
                self.n_bytes_read_since_last_heartbeat += len(msg)
            self.last_message = msg
        # send the message to the _process_message function
        retval = self._process_message(self.lock, msg)
        # count and (asynchronously) commit the message as processed
        if retval:
            with self.lock:
                self.n_msgs_processed += 1
                self.n_msgs_processed_since_last_heartbeat += 1
                if (
                    hasattr(msg, "key")
                    and hasattr(msg, "value")
                    and (
                        isinstance(msg.key, KafkaCryptoMessage)
                        or isinstance(msg.value, KafkaCryptoMessage)
                    )
                ):
                    self.n_bytes_processed_since_last_heartbeat += len(bytes(msg))
                elif isinstance(msg, Message):
                    self.n_bytes_processed_since_last_heartbeat += len(msg.value)
                else:
                    self.n_bytes_processed_since_last_heartbeat += len(msg)
            if not consumer.message_consumed_before(msg):
                tps = consumer.commit(msg)
                if tps is None:
                    return
                for t_p in tps:
                    if t_p.error is not None:
                        warnmsg = (
                            f"WARNING: failed to synchronously commit offset of last message "
                            f'received on "{t_p.topic}" partition {t_p.partition}. '
                            "Duplicate messages may result when this Consumer is restarted. "
                            f"Error reason: {t_p.error.str()}"
                        )
                        self.logger.warning(warnmsg)

    def __commit_last_message_offset(self, consumer):
        """
        Commit the offset of the last message received if it wasn't already consumed in the past
        (block until done)
        """
        try:
            tps = consumer.commit(self.last_message, asynchronous=False)
            if tps is not None:
                for t_p in tps:
                    if t_p.error is not None:
                        warnmsg = (
                            f"WARNING: failed to synchronously commit offset of last message "
                            f'received on "{t_p.topic}" partition {t_p.partition}. '
                            "Duplicate messages may result when this Consumer is restarted. "
                            "Error reason: {t_p.error.str()}"
                        )
                        self.logger.warning(warnmsg)
        except Exception as exc:
            warnmsg = (
                "WARNING: failed to synchronously commit offset of last message received. "
                "Duplicate messages may be read the next time this Consumer is started. "
                "Error will be logged below but not re-raised."
            )
            self.logger.warning(warnmsg, exc_info=exc)

    @classmethod
    def get_command_line_arguments(cls):
        superargs, superkwargs = super().get_command_line_arguments()
        args = [
            *superargs,
            "download_regex",
        ]
        return args, superkwargs

    @classmethod
    def get_init_args_kwargs(cls, parsed_args):
        superargs, superkwargs = super().get_init_args_kwargs(parsed_args)
        kwargs = {
            **superkwargs,
            "filepath_regex": parsed_args.download_regex,
        }
        return superargs, kwargs



"""Various types of DataFiles that have been read back from messages in a topic"""

# imports
import pathlib, os
from hashlib import sha512
from contextlib import nullcontext
from abc import ABC, abstractmethod
from ..config import DATA_FILE_HANDLING_CONST
from .data_file import DataFile


class DownloadDataFile(DataFile, ABC):
    """
    Class to represent a data file that will be read as messages from a topic

    :param filepath: Path to the file
    :type filepath: :class:`pathlib.Path`
    """

    #################### PROPERTIES AND STATIC METHODS ####################

    @staticmethod
    def get_full_filepath(dfc):
        """
        Return the full filepath of a downloaded file given one of its DataFileChunks

        :param dfc: One of the DataFileChunk objects contributing to the file
        :type dfc: :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk`

        :return: the full path to the file
        :rtype: :class:`pathlib.Path`
        """
        if dfc.filename_append == "":
            return dfc.filepath
        return dfc.filepath.with_name(
            dfc.filepath.stem + dfc.filename_append + dfc.filepath.suffix
        )

    @property
    @abstractmethod
    def bytestring(self):
        """
        A bytestring of the data contained in the file. Not implemented in the base class.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def check_file_hash(self):
        """
        The hash of the data in the file after it was read. Not implemented in the base class.
        """
        raise NotImplementedError

    @property
    def relative_filepath(self):
        """
        The path to the file, relative to its root directory
        """
        if self.subdir_str == "":
            return pathlib.Path(self.filename)
        return pathlib.Path(self.subdir_str + "/" + self.filename)

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self, filepath, *args, **kwargs):
        super().__init__(filepath, *args, **kwargs)
        # start an empty set of this file's downloaded offsets
        self._chunk_offsets_downloaded = []
        self.full_filepath = None
        self.subdir_str = None
        self.n_total_chunks = None

    def add_chunk(self, dfc, thread_lock=nullcontext()):
        """
        Process a chunk that's been read from a topic.
        Returns a number of codes based on what effect adding the chunk had.

        This function calls :func:`~_on_add_chunk`,
        with the :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk` as the argument.

        :param dfc: the DataFileChunk object whose data should be added
        :type dfc: :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk`
        :param thread_lock: the lock object to acquire/release so that race conditions
            don't affect reconstruction of the files
            (only needed if running this function asynchronously)
        :type thread_lock: :class:`threading.Lock`, optional

        :return: an internal code indicating whether the chunk: was successfully added to
            a file in progress, was already received, was the last chunk needed and
            the file is successfully reconstructed according to its hash or the
            post-reconstruction hash is mismatched to the hash of the file contents
            originally read from disk.
        :rtype: int
        """
        # if this chunk's offset has already been written to disk, return the "already written" code
        with thread_lock:
            already_written = dfc.chunk_offset_write in self._chunk_offsets_downloaded
        if already_written:
            return DATA_FILE_HANDLING_CONST.CHUNK_ALREADY_WRITTEN_CODE
        # the filepath of this DownloadDataFile and of the given DataFileChunk must match
        if dfc.filepath != self.filepath:
            errmsg = (
                f"ERROR: filepath mismatch between data file chunk with {dfc.filepath} "
                f"and data file with {self.filepath}"
            )
            self.logger.error(errmsg, exc_type=ValueError)
        # modify the filepath to include any append to the name
        full_filepath = self.__class__.get_full_filepath(dfc)
        if self.full_filepath is None:
            self.full_filepath = full_filepath
            self.filename = self.full_filepath.name
        elif self.full_filepath != full_filepath:
            errmsg = (
                f"ERROR: filepath for data file chunk {dfc.chunk_i}/{dfc.n_total_chunks} "
                f"with offset {dfc.chunk_offset_write} is {full_filepath} but the file "
                f"being reconstructed is expected to have filepath {self.full_filepath}"
            )
            self.logger.error(errmsg, exc_type=ValueError)
        # add the subdirectory string to this file
        if self.subdir_str is None:
            self.subdir_str = dfc.subdir_str
        elif self.subdir_str != dfc.subdir_str:
            errmsg = (
                f"Mismatched subdirectory strings! From file = {self.subdir_str}, "
                f"from chunk = {dfc.subdir_str}"
            )
            self.logger.error(errmsg, exc_type=ValueError)
        # set or check the total number of chunks expected
        if self.n_total_chunks is None:
            with thread_lock:
                self.n_total_chunks = dfc.n_total_chunks
        elif self.n_total_chunks != dfc.n_total_chunks:
            errmsg = (
                f"ERROR: {self.__class__.__name__} with filepath {self.full_filepath} "
                f"is expecting {self.n_total_chunks} chunks but found a chunk from a split "
                f"with {dfc.n_total_chunks} total chunks."
            )
            self.logger.error(errmsg, exc_type=ValueError)
        with thread_lock:
            # call the function to actually add the chunk
            self._on_add_chunk(dfc)
            # add the offset of the added chunk to the set of reconstructed file chunks
            self._chunk_offsets_downloaded.append(dfc.chunk_offset_write)
            last_chunk = len(self._chunk_offsets_downloaded) == dfc.n_total_chunks
        # if this chunk was the last that needed to be added, check the hashes
        if last_chunk:
            if self.check_file_hash != dfc.file_hash:
                return DATA_FILE_HANDLING_CONST.FILE_HASH_MISMATCH_CODE
            return DATA_FILE_HANDLING_CONST.FILE_SUCCESSFULLY_RECONSTRUCTED_CODE
        return DATA_FILE_HANDLING_CONST.FILE_IN_PROGRESS

    #################### PRIVATE HELPER FUNCTIONS ####################

    @abstractmethod
    def _on_add_chunk(self, dfc):
        """
        A function to actually process a new chunk being added to the file.
        This function is executed while a thread lock is acquired so it will never
        run asynchronously. Also any DataFileChunks passed to this function are
        guaranteed to have unique offsets.

        Not implemented in the base class.

        :param dfc: the DataFileChunk object whose data should be added
        :type dfc: :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk`
        """
        raise NotImplementedError


class DownloadDataFileToDisk(DownloadDataFile):
    """
    Class to represent a data file that will be reconstructed on disk
    using messages read from a topic

    :param filepath: Path to the file
    :type filepath: :class:`pathlib.Path`
    """

    #################### CONSTANTS ####################

    READ_BUFFER_SIZE = 4096

    #################### PROPERTIES ####################

    @property
    def bytestring(self):
        """
        A bytestring of the data in the file
        """
        with open(self.full_filepath, "rb") as fp:
            data = fp.read()
        return data

    @property
    def check_file_hash(self):
        """
        Hash of the file contents as read from its current location on disk
        """
        check_file_hash = sha512()
        with open(self.full_filepath, "rb") as fp:
            while True:
                data = fp.read(self.READ_BUFFER_SIZE)
                if not data:
                    break
                check_file_hash.update(data)
        return check_file_hash.digest()

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self, filepath, *args, **kwargs):
        super().__init__(filepath, *args, **kwargs)
        # create the parent directory of the file if it doesn't exist yet
        if not self.filepath.parent.is_dir():
            self.filepath.parent.mkdir(parents=True)

    def _on_add_chunk(self, dfc):
        """
        Add the data from a given file chunk to this file on disk

        :param dfc: the DataFileChunk object whose data should be added
        :type dfc: :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk`
        """
        try:
            super()._on_add_chunk(dfc)
        except NotImplementedError:
            pass
        mode = "r+b" if self.full_filepath.is_file() else "w+b"
        with open(self.full_filepath, mode) as fp:
            fp.seek(dfc.chunk_offset_write)
            fp.write(dfc.data)
            fp.flush()
            os.fsync(fp.fileno())
            fp.close()


class DownloadDataFileToMemory(DownloadDataFile):
    """
    Class to represent a data file that will be held in memory
    and populated by the contents of messages from a topic

    :param filepath: Path to the file
    :type filepath: :class:`pathlib.Path`
    """

    #################### PROPERTIES ####################

    @property
    def bytestring(self):
        """
        The bytestring of the file contents (like calling file_pointer.read() for a file
        opened in "rb" mode). Call bytestring.decode() to convert this to a text string.
        """
        if self.__bytestring is None:
            self.__create_bytestring()
        return self.__bytestring

    @bytestring.setter
    def bytestring(self, new_bytestring):
        if self.__bytestring is not None:
            warnmsg = (
                f"WARNING: resetting a non-None bytestring for {self.__class__.__name__} "
                f"with path {self.filepath}"
            )
            self.logger.warning(warnmsg)
        self.__bytestring = new_bytestring

    @property
    def check_file_hash(self):
        """
        The hash of the file contents determined from the bytestring of the file stored in memory
        """
        check_file_hash = sha512()
        check_file_hash.update(self.bytestring)
        return check_file_hash.digest()

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self, filepath, *args, **kwargs):
        super().__init__(filepath, *args, **kwargs)
        # start a dictionary of the file data by their offsets
        self.__chunk_data_by_offset = {}
        # placeholder for the eventual full data bytestring
        self.__bytestring = None

    #################### PRIVATE HELPER FUNCTIONS ####################

    def _on_add_chunk(self, dfc):
        """
        Add the data from a given file chunk to the dictionary of data by offset

        :param dfc: the DataFileChunk object whose data should be added
        :type dfc: :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk`
        """
        try:
            super()._on_add_chunk(dfc)
        except NotImplementedError:
            pass
        self.__chunk_data_by_offset[dfc.chunk_offset_write] = dfc.data

    def __create_bytestring(self):
        """
        Makes all of the data held in the dictionary into a single bytestring
        """
        bytestring = b""
        for data in [
            self.__chunk_data_by_offset[offset]
            for offset in sorted(self.__chunk_data_by_offset.keys())
        ]:
            bytestring += data
        self.__bytestring = bytestring


class DownloadDataFileToMemoryAndDisk(DownloadDataFileToMemory, DownloadDataFileToDisk):
    """
    A file that's written to disk as its messages are read
    from a topic, while also having a bytestring created for it.

    :param filepath: Path to the file
    :type filepath: :class:`pathlib.Path`
    """

    @property
    def check_file_hash(self):
        """
        Hash of the file contents as read from its current location on disk
        and from the bytestring of the file stored in memory.
        Raises an error if they don't match.
        """
        check_file_hash_memory = sha512()
        check_file_hash_memory.update(self.bytestring)
        check_file_hash_memory = check_file_hash_memory.digest()
        check_file_hash_disk = sha512()
        with open(self.full_filepath, "rb") as fp:
            while True:
                data = fp.read(self.READ_BUFFER_SIZE)
                if not data:
                    break
                check_file_hash_disk.update(data)
        check_file_hash_disk = check_file_hash_disk.digest()
        if check_file_hash_disk != check_file_hash_memory:
            errmsg = (
                f"ERROR: hashes of file on disk and in memory are mismatched! "
                f"On disk: {check_file_hash_disk}. In memory: {check_file_hash_memory}."
            )
            self.logger.error(errmsg, exc_type=ValueError)
        return check_file_hash_disk

    def __init__(self, filepath, *args, **kwargs):
        super().__init__(filepath, *args, **kwargs)



"""Custom argument parser and associated functions"""

# imports
import pathlib, re
from openmsitoolbox import OpenMSIArgumentParser
from openmsitoolbox.argument_parsing.parser_callbacks import (
    existing_dir,
    positive_int,
    int_power_of_two,
)
from .config import RUN_CONST

#################### CALLBACK FUNCTIONS ####################


def config_path(configarg):
    """
    convert a string or path argument into a config file path
    (raise an exception if the file can't be found)
    """
    if isinstance(configarg, str) and "." not in configarg:
        configarg += RUN_CONST.CONFIG_FILE_EXT
    configpath = pathlib.Path(configarg)
    if configpath.is_file():
        return configpath.resolve()
    if (RUN_CONST.CONFIG_FILE_DIR / configpath).is_file():
        return (RUN_CONST.CONFIG_FILE_DIR / configpath).resolve()
    raise ValueError(
        f"ERROR: config argument {configarg} is not a recognized config file!"
    )


def detect_bucket_name(argstring):
    """
    detects if the bucket name contains invalid characters
    """
    if argstring is None:  # Then the argument wasn't given and nothing should be done
        return None
    illegal_characters = [
        "#",
        "%",
        "&",
        "{",
        "}",
        "\\",
        "/",
        "<",
        ">",
        "*",
        "?",
        " ",
        "$",
        "!",
        "'",
        '"',
        ":",
        "@",
        "+",
        "`",
        "|",
        "=",
    ]
    for illegal_character in illegal_characters:
        if illegal_character in argstring:
            raise RuntimeError(f"ERROR: Illegal characters in bucket_name {argstring}")
    return argstring


def positive_float(argval):
    """
    make sure a given value is a positive float
    """
    argval = float(argval)
    if (not isinstance(argval, float)) or (argval < 0):
        raise ValueError(f"ERROR: invalid argument: {argval} must be a positive float!")
    return argval


#################### MYARGUMENTPARSER CLASS ####################


class OpenMSIStreamArgumentParser(OpenMSIArgumentParser):
    """
    An ArgumentParser with some commonly-used arguments in it.

    All constructor arguments get passed to the underlying :class:`argparse.ArgumentParser` object.

    Arguments for the parser are defined in the :attr:`~OpenMSIStreamArgumentParser.ARGUMENTS`
    class variable, which is a dictionary. The keys are names of arguments, and the values
    are lists. The first entry in each list is a string reading "positional" or "optional"
    depending on the type of argument, and the second entry is a dictionary of keyword arguments
    to send to :func:`argparse.ArgumentParser.add_argument`.
    """

    DEF_HEARTBEAT_INTERVAL = 15 * 60  # send heartbeats every 15 minutes by default

    ARGUMENTS = {
        **OpenMSIArgumentParser.ARGUMENTS,
        "upload_dir": [
            "positional",
            {
                "type": existing_dir,
                "help": "Path to the directory to watch for files to upload",
            },
        ],
        "bucket_name": [
            "positional",
            {
                "type": detect_bucket_name,
                "help": (
                    "Name of the object store bucket to which consumed files should be "
                    "transferred. Name should only contain valid characters"
                ),
            },
        ],
        "config": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_CONFIG_FILE,
                "type": config_path,
                "help": (
                    f"Name of config file to use in {RUN_CONST.CONFIG_FILE_DIR.resolve()}, "
                    "or path to a file in a different location"
                ),
            },
        ],
        "topic_name": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_TOPIC_NAME,
                "help": "Name of the topic to produce to or consume from",
            },
        ],
        "consumer_topic_name": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_TOPIC_NAME,
                "help": "Name of the topic to consume from",
            },
        ],
        "producer_topic_name": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_TOPIC_NAME,
                "help": "Name of the topic to produce to",
            },
        ],
        "n_consumer_threads": [
            "optional",
            {
                "default": RUN_CONST.N_DEFAULT_DOWNLOAD_THREADS,
                "type": positive_int,
                "help": "Number of consumer threads to use",
            },
        ],
        "n_producer_threads": [
            "optional",
            {
                "default": RUN_CONST.N_DEFAULT_UPLOAD_THREADS,
                "type": positive_int,
                "help": "Number of producer threads to use",
            },
        ],
        "upload_regex": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_UPLOAD_REGEX,
                "type": re.compile,
                "help": "Only files with paths matching this regular expression will be uploaded",
            },
        ],
        "download_regex": [
            "optional",
            {
                "type": re.compile,
                "help": "Only files with paths matching this regular expression will be uploaded",
            },
        ],
        "chunk_size": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_CHUNK_SIZE,
                "type": int_power_of_two,
                "help": (
                    "Max size (in bytes) of chunks into which files should be broken "
                    "as they are uploaded"
                ),
            },
        ],
        "queue_max_size": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_MAX_UPLOAD_QUEUE_MEGABYTES,
                "type": positive_int,
                "help": (
                    "Maximum allowed size in MB of the internal upload queue. "
                    "Use to adjust RAM usage if necessary."
                ),
            },
        ],
        "mode": [
            "optional",
            {
                "choices": ["memory", "disk", "both"],
                "default": "memory",
                "help": (
                    "Choose whether files should be reconstructed in 'memory', "
                    "on 'disk' (with the output directory as the root directory) "
                    "or 'both' (giving access to the data bytestring, "
                    "but also writing to disk)"
                ),
            },
        ],
        "upload_existing": [
            "optional",
            {
                "action": "store_true",
                "help": (
                    "Add this flag to upload files already existing in addition to those "
                    "added to the directory after this code starts running (by default "
                    "only files added after startup will be uploaded)"
                ),
            },
        ],
        "watchdog_lag_time": [
            "optional",
            {
                "default": RUN_CONST.DEFAULT_WATCHDOG_LAG_TIME,
                "type": int,
                "help": (
                    "Number of seconds that a file must remain static (unmodified) "
                    "for its upload to begin"
                ),
            },
        ],
        "use_polling_observer": [
            "optional",
            {
                "action": "store_true",
                "help": (
                    "Add this flag to use a PollingObserver to trigger watchdog events "
                    "instead of the default Observer. This is recommended when reading "
                    "files from CIFS/SMB mounted directories, or if the default Observer "
                    "causes any problems in a deployment."
                ),
            },
        ],
        "consumer_group_id": [
            "optional",
            {"default": "create_new", "help": "ID to use for all consumers in the group"},
        ],
        "treat_undecryptable_as_plaintext": [
            "optional",
            {
                "action": "store_true",
                "help": (
                    "If this flag is present, Consumer(s) will return message "
                    "keys/values that will never be decryptable as plaintext. "
                    "This speeds up processing of messages that will never be "
                    "decryptable. For use in cases like enabling/disabling encryption "
                    "across a platform, or when encrypted and unencrypted messages are "
                    "mixed in the same topic."
                ),
            },
        ],
        "girder_api_url": [
            "positional",
            {
                "help": (
                    "The full path to the REST API of a Girder instance, "
                    "e.g. http://my.girder.com/api/v1."
                )
            },
        ],
        "girder_api_key": [
            "positional",
            {"help": "The API key to use for authenticating to the Girder instance"},
        ],
        "girder_root_folder_id": [
            "optional",
            {
                "help": (
                    "The ID of the Girder Folder relative to which files should be "
                    "uploaded. If this argument is given, it will supersede both of the "
                    "'collection_name' and 'girder_root_folder_path' arguments."
                ),
            },
        ],
        "collection_name": [
            "optional",
            {
                "help": (
                    "The name of the top-level Collection to which files should be uploaded. "
                    "Superseded if 'girder_root_folder_id' is given."
                ),
            },
        ],
        "girder_root_folder_path": [
            "optional",
            {
                "help": (
                    "The name of the Folder inside the top-level Collection relative to "
                    "which files should be uploaded. A path to a subdirectory in the "
                    "Collection can be given using forward slashes. "
                    "Superseded if 'girder_root_folder_id' is given. "
                    "(default = collection_name/topic_name)"
                ),
            },
        ],
        "metadata": [
            "optional",
            {
                "help": (
                    "Adding this argument will add a corresponding metadata "
                    "field to all created folders and items. Assumes JSON serialized string."
                ),
            },
        ],
        "heartbeat_topic_name": [
            "optional",
            {
                "help": (
                    "Name of the topic to which heartbeat messages for the long-running "
                    "program should be produced. This argument must be included to "
                    "produce heartbeat messages"
                ),
            },
        ],
        "heartbeat_program_id": [
            "optional",
            {
                "help": (
                    "ID to include in keys of heartbeat messages to uniquely identify "
                    "this particular long-running program instance. (Default is the hex "
                    "code of the heartbeat producer's address in memory which is NOT "
                    "static)"
                ),
            },
        ],
        "heartbeat_interval_secs": [
            "optional",
            {
                "default": DEF_HEARTBEAT_INTERVAL,
                "type": positive_float,
                "help": (
                    "How often (in seconds) messages should be produced to the "
                    "heartbeat topic configured for the long-running program"
                ),
            },
        ],
    }



"""Utilities for working with S3 buckets"""

import hashlib
import boto3
from openmsitoolbox import LogOwner


class S3Service(LogOwner):
    """
    Wrapper around an S3 connection session
    """

    def __init__(self, s3_config, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = boto3.session.Session()

        self.bucket_name = s3_config["bucket_name"]
        endpoint_url = str(s3_config["endpoint_url"])

        if not endpoint_url.startswith("https://"):
            endpoint_url = "https://" + endpoint_url
        self.s3_client = self.session.client(
            service_name="s3",
            aws_access_key_id=s3_config["access_key_id"],
            aws_secret_access_key=s3_config["secret_key_id"],
            region_name=s3_config["region"],
            endpoint_url=endpoint_url,
        )
        self.region = s3_config["region"]
        self.grant_read = 'uri="http://acs.amazonaws.com/groups/global/AllUsers"'

    def get_object_stream_by_object_key(self, bucket_name, object_key):
        """
        Return the stream of an object in a bucket given its key
        """
        s3_response_object = self.s3_client.get_object(Bucket=bucket_name, Key=object_key)
        return s3_response_object["Body"].read()

    def delete_object_from_bucket(self, bucket_name, object_key):
        """
        Delete an object with a given key from a bucket
        """
        try:
            # make sure object exits before deleting the object from the bucket
            s3_response_object = self.s3_client.get_object(
                Bucket=bucket_name, Key=object_key
            )
            if s3_response_object is None:
                msg = (
                    "The object "
                    + object_key
                    + " does not exist in this bucket: "
                    + bucket_name
                )
                self.logger.error(msg)
                return
        except Exception:
            msg = (
                "A problem occurred while reading "
                + object_key
                + " from bucket: "
                + bucket_name
            )
            self.logger.error(msg)
            return
        # delete the object from the bucket safely
        try:
            self.s3_client.delete_object(Bucket=bucket_name, Key=object_key)
            msg = object_key + " was deleted successfully from bucket: " + bucket_name
            self.logger.debug(msg)
        except Exception:
            msg = "Could not delete " + object_key + " from bucket: " + bucket_name
            self.logger.error(msg)
            return

    def compare_producer_datafile_with_s3_object_stream(
        self, bucket_name, object_key, hashed_datafile_stream
    ):
        """
        Compare a given DataFile hash with the hash of an object in a bucket
        """
        if hashed_datafile_stream is None:
            return False
        s3_response_object = self.s3_client.get_object(Bucket=bucket_name, Key=object_key)
        object_content = s3_response_object["Body"].read()
        md5 = hashlib.md5()
        md5.update(object_content)
        if object_content is None:
            return False
        hhh = format(md5.hexdigest())
        return str(hhh) == hashed_datafile_stream

    def compare_consumer_datafile_with_s3_object_stream(
        self, bucket_name, object_key, datafile
    ):
        """
        Compare a given DataFile with an object in a bucket using their hashes
        """
        if datafile is None:
            return False
        s3_response_object = self.s3_client.get_object(Bucket=bucket_name, Key=object_key)
        object_content = s3_response_object["Body"].read()
        if object_content is None:
            return False
        datafile_md5 = hashlib.md5()
        datafile_md5.update(datafile.bytestring)
        object_content_md5 = hashlib.md5()
        object_content_md5.update(object_content)
        return format(datafile_md5.hexdigest()) == format(object_content_md5.hexdigest())



"""Constants used in running different controlled processes in general"""

# imports
import os, pathlib, re


class RunConstants:
    """
    Constants used for running controlled processes, reading config files, etc.
    """

    CONFIG_FILE_EXT = ".config"
    CONFIG_FILE_DIR = (
        os.environ["OPENMSISTREAM_CONFIG_FILE_DIR"]
        if "OPENMSISTREAM_CONFIG_FILE_DIR" in os.environ
        else (
            pathlib.Path(__file__).parent.parent / "kafka_wrapper" / "config_files"
        ).resolve()
    )
    # name of the config file that will be used by default
    DEFAULT_CONFIG_FILE = "test"
    # name of the topic to produce to by default
    DEFAULT_TOPIC_NAME = "test"
    # default number of threads to use when uploading a file
    N_DEFAULT_UPLOAD_THREADS = 2
    # default number of threads to use when downloading chunks of a file
    N_DEFAULT_DOWNLOAD_THREADS = 2
    # name of the config file used in "real" production
    PRODUCTION_CONFIG_FILE = "prod"
    # matches everything except something starting with a '.' or ending in '.log'
    DEFAULT_UPLOAD_REGEX = re.compile(r"^((?!(\.|.*.log))).*$")
    # default size in bytes of each file upload chunk
    DEFAULT_CHUNK_SIZE = 524288
    # default maximum size (in MB) of the internal upload Queue
    DEFAULT_MAX_UPLOAD_QUEUE_MEGABYTES = 500
    # default number of seconds that a file must remain static for uploading to start
    DEFAULT_WATCHDOG_LAG_TIME = 3
    # default Girder collection name
    DEFAULT_COLLECTION_NAME = "WholeTale Catalog"


RUN_CONST = RunConstants()



"""Wrapper around a Python ConfigParser to simplify some commonly-used operations"""

# imports
import os, pathlib, configparser
from openmsitoolbox import LogOwner


class ConfigFileParser(LogOwner):
    """
    Parses configurations from files
    """

    #################### PROPERTIES ####################

    @property
    def has_default(self):
        """
        True if a config file has a DEFAULT section
        """
        return "DEFAULT" in self._config

    @property
    def available_group_names(self):
        """
        The list of section names in the config file
        """
        return self._config.sections()

    @property
    def env_var_names(self):
        """
        A generator of the environment variable names used in the config file
        """
        for csd in self._config.values():
            for v in csd.values():
                if v.startswith("$"):
                    yield v[1:]

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self, config_path, *args, **kwargs):
        """
        config_path = path to the config file to parse
        """
        super().__init__(*args, **kwargs)
        if isinstance(config_path, str):
            config_path = pathlib.Path(config_path)
        self.filepath = config_path
        if not config_path.is_file():
            self.logger.error(
                f"ERROR: configuration file {config_path} does not exist!",
                exc_type=FileNotFoundError,
            )
        self._config = configparser.ConfigParser()
        self._config.read(config_path)

    def get_config_dict_for_groups(self, group_names):
        """
        Return a config dictionary populated with configurations from groups with the
        given names

        group_names = the list of group names to add to the dictionary (or one string)
        """
        if isinstance(group_names, str):
            group_names = [group_names]
        config_dict = {}
        for group_name in group_names:
            if group_name not in self._config:
                errmsg = (
                    f"ERROR: {group_name} is not a recognized section in {self.filepath}!"
                )
                self.logger.error(errmsg, exc_type=ValueError)
            for key, value in self._config[group_name].items():
                # don't add the 'node_id' to groups for brokers, producers, or consumers
                if key == "node_id" and any(
                    stem in group_name for stem in ("broker", "producer", "consumer")
                ):
                    continue
                # if the value is an environment variable, expand it on the current system
                if value.startswith("$"):
                    exp_value = os.path.expandvars(value)
                    if exp_value == value:
                        errmsg = (
                            f"ERROR: Expanding {value} in {self.filepath} as an "
                            "environment variable failed (must be set on system)"
                        )
                        self.logger.error(errmsg, exc_type=ValueError)
                    else:
                        value = exp_value
                config_dict[key] = value
        return config_dict

    #################### PRIVATE HELPER FUNCTIONS ####################

    def _get_config_dict(self, group_name):
        to_return = {}
        if group_name in self.available_group_names:
            to_return = self.get_config_dict_for_groups(group_name)
        return to_return



"""
Code for managing the two csv files listing files that are
in the process of being produced/have been fully produced
"""

# imports
import pathlib, datetime
from typing import Set
from dataclasses import dataclass
from openmsitoolbox import LogOwner
from ....utilities import (
    DataclassTableReadOnly,
    DataclassTableAppendOnly,
    DataclassTable,
)


@dataclass
class RegistryLineInProgress:
    """
    A line in the table listing files in progress
    """

    filename: str
    rel_filepath: pathlib.Path
    n_chunks: int
    n_chunks_delivered: int
    n_chunks_to_send: int
    started: datetime.datetime
    chunks_delivered: Set[int]
    chunks_to_send: Set[int]


@dataclass
class RegistryLineCompleted:
    """
    A line in the table listing completed files
    """

    filename: str
    rel_filepath: pathlib.Path
    n_chunks: int
    started: datetime.datetime
    completed: datetime.datetime


class ProducerFileRegistry(LogOwner):
    """
    Manages two atomic csv files listing which portions
    of which files have been uploaded to a particular topic
    """

    def __init__(
        self, dirpath, topic_name, *args, max_completed_lines_per_file=1000, **kwargs
    ):
        """
        dirpath    = path to the directory that should contain the csv file
        topic_name = the name of the topic that will be produced to (used in the filename)
        """
        super().__init__(*args, **kwargs)
        # A table to list the files currently in progress
        in_prog_filepath = dirpath / f"upload_to_{topic_name}_in_progress.csv"
        self.__in_prog = DataclassTable(
            dataclass_type=RegistryLineInProgress,
            filepath=in_prog_filepath,
            logger=self.logger,
        )
        # Pattern for the "completed" .csv file names
        self.completed_filepath_pattern = dirpath / f"uploaded_to_{topic_name}.csv"
        # keep track of the maximum number of lines per "completed" file
        self.max_completed_lines_per_file = max_completed_lines_per_file
        # Make a dictionary to hold tables of any newly received "completed" entries
        self.__completed_tables_by_path = {}
        # Consolidate existing "completed" csv files into one
        self.consolidate_completed_files()

    def __del__(self):
        self.consolidate_completed_files()

    def consolidate_completed_files(self):
        """
        Search the directory holding the files for all "completed" file entries
        and write them into a single file
        """
        # add any lines from files in the directory to the one consolidated file
        stem = self.completed_filepath_pattern.stem
        suffix = self.completed_filepath_pattern.suffix
        globpattern = f"{stem}*{suffix}"
        consolidated_file = None
        for fp in self.completed_filepath_pattern.parent.glob(globpattern):
            if fp == self.completed_filepath_pattern:
                continue
            if fp in self.__completed_tables_by_path:
                self.__completed_tables_by_path[fp].dump_to_file()
            file_to_add = DataclassTableReadOnly(
                dataclass_type=RegistryLineCompleted, filepath=fp, logger=self.logger
            )
            if consolidated_file is None:
                consolidated_file = DataclassTableAppendOnly(
                    dataclass_type=RegistryLineCompleted,
                    filepath=self.completed_filepath_pattern,
                    logger=self.logger,
                )
            consolidated_file.add_entries(file_to_add.objects)
        if consolidated_file is None:
            return
        # dump out the consolidated file
        consolidated_file.dump_to_file()
        # make sure that all the lines from the individual files have been successfully copied
        all_lines = (
            DataclassTableReadOnly(
                dataclass_type=RegistryLineCompleted,
                filepath=self.completed_filepath_pattern,
                logger=self.logger,
            )
        ).lines
        for fp in self.completed_filepath_pattern.parent.glob(globpattern):
            if fp == self.completed_filepath_pattern:
                continue
            added_file = DataclassTableReadOnly(
                dataclass_type=RegistryLineCompleted, filepath=fp, logger=self.logger
            )
            file_is_good = True
            for entry_line in added_file.lines:
                if entry_line not in all_lines:
                    self.logger.warning(f"missing line = {entry_line} in file {fp}")
                    warnmsg = (
                        "WARNING: failed to consolidate individual files into "
                        f"{self.completed_filepath_pattern}. Individual files will be "
                        "retained and should be manually concatenated. "
                        "Duplicate entries may be present."
                    )
                    self.logger.warning(warnmsg)
                    file_is_good = False
            if file_is_good:
                if fp in self.__completed_tables_by_path:
                    self.__completed_tables_by_path.pop(fp)
                fp.unlink()

    def get_incomplete_filepaths_and_chunks(self):
        """
        Generate tuples of (rel_filepath, chunks to upload) for each file
        that has not yet been completely uploaded
        """
        for obj_address in self.__in_prog.obj_addresses:
            attr_dict = self.__in_prog.get_entry_attrs(
                obj_address, "rel_filepath", "chunks_to_send"
            )
            if len(attr_dict["chunks_to_send"]) > 0:
                yield attr_dict["rel_filepath"], attr_dict["chunks_to_send"]

    def get_completed_filepaths(self):
        """
        Generate relative filepaths for each file that has been completely uploaded
        """
        self.consolidate_completed_files()
        completed = DataclassTableReadOnly(
            dataclass_type=RegistryLineCompleted,
            filepath=self.completed_filepath_pattern,
            logger=self.logger,
        )
        seen = set()
        for obj_address in completed.obj_addresses:
            rel_filepath = completed.get_entry_attrs(obj_address, "rel_filepath")
            if rel_filepath not in seen:
                seen.add(rel_filepath)
                yield rel_filepath

    def register_chunk(self, filename, rel_filepath, n_total_chunks, chunk_i, prodid):
        """
        Register a chunk as having been successfully produced by a particular producer
        Returns True if all chunks for the file have been produced
        """
        # acquire the lock so no other threads can change the in progress file
        self.__in_prog.lock.acquire()
        # get a dictionary of the existing object addresses keyed by their filepaths
        existing_obj_addresses = self.__in_prog.obj_addresses_by_key_attr("rel_filepath")
        # if the file is already recognized as in progress
        if rel_filepath in existing_obj_addresses:
            return self.__register_chunk_of_existing_file(
                filename, rel_filepath, n_total_chunks, chunk_i, prodid
            )
        # otherwise it's a new file to list somewhere
        return self.__register_chunk_for_new_file(
            filename, rel_filepath, n_total_chunks, chunk_i, prodid
        )

    def __register_chunk_of_existing_file(
        self, filename, rel_filepath, n_total_chunks, chunk_i, prodid
    ):
        """
        Register a chunk that belongs to an existing file.
        Returns True if all chunks for the file have been received by the broker.
        """
        # get a dictionary of the existing object addresses keyed by their filepaths
        existing_obj_addresses = self.__in_prog.obj_addresses_by_key_attr("rel_filepath")
        # make sure there's only one object with this filepath
        if len(existing_obj_addresses[rel_filepath]) != 1:
            errmsg = (
                f"ERROR: found {len(existing_obj_addresses[rel_filepath])} files in the "
                f"producer registry for filepath {rel_filepath}"
            )
            self.__in_prog.lock.release()
            self.logger.error(errmsg, exc_type=RuntimeError)
        existing_addr = existing_obj_addresses[rel_filepath][0]
        # make sure the total numbers of chunks match
        existing_n_chunks = self.__in_prog.get_entry_attrs(existing_addr, "n_chunks")
        if existing_n_chunks != n_total_chunks:
            errmsg = (
                f"ERROR: {rel_filepath} in {self.__class__.__name__} is listed as having "
                f"{existing_n_chunks} total chunks, but the producer callback for this "
                f"file lists {n_total_chunks} chunks! Did the chunk size change?"
            )
            self.__in_prog.lock.release()
            self.logger.error(errmsg, exc_type=RuntimeError)
        # get its current state
        attrs = self.__in_prog.get_entry_attrs(
            existing_addr, "chunks_delivered", "chunks_to_send"
        )
        # if the chunk is already registered, just return
        if chunk_i in attrs["chunks_delivered"]:
            self.__in_prog.lock.release()
            return False
        # otherwise add/remove it from the sets
        attrs["chunks_delivered"].add(chunk_i)
        try:
            attrs["chunks_to_send"].remove(chunk_i)
        except KeyError:
            pass
        # reset the attributes for the object
        new_attrs = {
            "n_chunks_delivered": len(attrs["chunks_delivered"]),
            "n_chunks_to_send": len(attrs["chunks_to_send"]),
            "chunks_delivered": attrs["chunks_delivered"],
            "chunks_to_send": attrs["chunks_to_send"],
        }
        self.__in_prog.set_entry_attrs(existing_addr, **new_attrs)
        # if there are no more chunks to send and all chunks have been delivered,
        # move this file from "in progress" to "completed" and force-dump the files
        if (
            self.__in_prog.get_entry_attrs(existing_addr, "n_chunks_delivered")
            == n_total_chunks
            and self.__in_prog.get_entry_attrs(existing_addr, "n_chunks_to_send") == 0
        ):
            started = self.__in_prog.get_entry_attrs(existing_addr, "started")
            completed_entry = RegistryLineCompleted(
                filename, rel_filepath, n_total_chunks, started, datetime.datetime.now()
            )
            self.__add_completed_entry(completed_entry, prodid)
            self.__in_prog.remove_entries(existing_addr)
            self.__in_prog.dump_to_file()
            self.__in_prog.lock.release()
            return True
        self.__in_prog.lock.release()
        return False

    def __register_chunk_for_new_file(
        self, filename, rel_filepath, n_total_chunks, chunk_i, prodid
    ):
        """
        Register the first chunk for a file.
        Returns True if the file only had one chunk.
        """
        to_deliver = set([])
        for ichunk in range(1, n_total_chunks + 1):
            if ichunk != chunk_i:
                to_deliver.add(ichunk)
        # if there are other chunks to deliver, register it to "in_progress"
        if len(to_deliver) > 0:
            in_prog_entry = RegistryLineInProgress(
                filename,
                rel_filepath,
                n_total_chunks,
                1,
                n_total_chunks - 1,
                datetime.datetime.now(),
                set([chunk_i]),
                to_deliver,
            )
            self.__in_prog.add_entries(in_prog_entry)
            self.__in_prog.dump_to_file()
            self.__in_prog.lock.release()
            return False
        # otherwise register it as a completed file
        self.__in_prog.lock.release()
        completed_entry = RegistryLineCompleted(
            filename,
            rel_filepath,
            n_total_chunks,
            datetime.datetime.now(),
            datetime.datetime.now(),
        )
        self.__add_completed_entry(completed_entry, prodid)
        return True

    def __add_completed_entry(self, completed_entry, prodid):
        """
        Add a given entry from a given produced as "completed", possibly juggling
        the files around if they're getting too large
        """
        # create the path to the table file that should be added to
        stem = self.completed_filepath_pattern.stem
        suffix = self.completed_filepath_pattern.suffix
        prod_name = f"{stem}_p{prodid}{suffix}"
        table_path = self.completed_filepath_pattern.with_name(prod_name)
        create_new_table_at_path = False
        # if the table already exists
        if table_path in self.__completed_tables_by_path:
            # if the table has the max # of entries or more entries already
            if (
                self.__completed_tables_by_path[table_path].n_entries
                >= self.max_completed_lines_per_file
            ):
                # dump it out to its file
                self.__completed_tables_by_path[table_path].dump_to_file()
                # rename the dumped file with a timestamp
                timestamp = str(datetime.datetime.now().timestamp()).replace(".", "_")
                new_path = table_path.with_name(
                    f"{table_path.stem}_{timestamp}{table_path.suffix}"
                )
                table_path.rename(new_path)
                # reset the table associated with the filepath for this producer
                create_new_table_at_path = True
        else:
            create_new_table_at_path = True
        if create_new_table_at_path:
            self.__completed_tables_by_path[table_path] = DataclassTableAppendOnly(
                RegistryLineCompleted, filepath=table_path, logger=self.logger
            )
        self.__completed_tables_by_path[table_path].add_entries(completed_entry)
        self.__completed_tables_by_path[table_path].dump_to_file()



" Example of a general script that can be installed as a Service/daemon "

# imports
import pathlib, datetime, time
from argparse import ArgumentParser


def main(args=None):
    """
    The main function that will be run as a Service/daemon
    """
    parser = ArgumentParser()
    parser.add_argument(
        "output_dir",
        type=pathlib.Path,
        help="Path to the directory where the test file should be created",
    )
    args = parser.parse_args(args)
    if not args.output_dir.is_dir():
        if args.output_dir.exists():
            raise ValueError(
                f"ERROR: given output directory {args.output_dir} exists but is not a directory!"
            )
        args.output_dir.mkdir(parents=True)
    test_file_name = "script_example_service_test.txt"
    if not (args.output_dir / test_file_name).is_file():
        with open(args.output_dir / test_file_name, "w") as fp:
            fp.write(
                "This file was created to test running a generic Python script as a Service/daemon"
            )
    with open(args.output_dir / test_file_name, "a") as fp:
        fp.write(
            f'\nService rerun {(datetime.datetime.now()).strftime("on %Y-%m-%d at %H:%M:%S")}'
        )
    time.sleep(30)


if __name__ == "__main__":
    main()



"""Various helper functions for working with Services/daemons"""

# imports
import pathlib, os, sys, platform
from subprocess import check_output, CalledProcessError
from .config import SERVICE_CONST


def get_os_name():
    """
    Return the name of the operating system the Service is being installed or running on
    """
    if platform.system() == "Windows":
        return "Windows"
    if platform.system() == "Linux":
        return "Linux"
    # MacOS is not supported
    if platform.system() == "Darwin":
        errmsg = "ERROR: Installing programs as Services is not supported on MacOS!"
        SERVICE_CONST.logger.error(errmsg, exc_type=NotImplementedError)
    # otherwise I don't know what happened
    else:
        errmsg = (
            f"ERROR: could not determine operating system from platform.system() "
            f'output "{platform.system()}"'
        )
        SERVICE_CONST.logger.error(errmsg, exc_type=ValueError)
    return None


def run_cmd_in_subprocess(args, *, logger=None, reraise=False, **other_kwargs):
    """
    run a command in a subprocess and return its result, printing and re-raising
    any exceptions it causes
    """
    if isinstance(args, str):
        args = [args]
    try:
        result = check_output(args, env=os.environ, **other_kwargs)
        return result
    except CalledProcessError as exc:
        errmsg = "ERROR: failed to run a command! "
        if exc.output is not None and exc.output.strip() != "":
            errmsg += f"\noutput:\n{exc.output.decode()}"
        if exc.stdout is not None and exc.stdout.strip() != "":
            errmsg += f"\nstdout:\n{exc.stdout.decode()}"
        if exc.stderr is not None and exc.stderr.strip() != "":
            errmsg += f"\nstderr:\n{exc.stderr.decode()}"
        if logger is not None:
            logger.error(errmsg, exc_info=exc, reraise=reraise)
        else:
            SERVICE_CONST.logger.error(errmsg, exc_info=exc, reraise=reraise)
        return None


def set_env_var(var_name, var_val):
    """
    set an environment variable given its name and value
    """
    if get_os_name() == "Windows":
        pwrsh_cmd = (
            f'[Environment]::SetEnvironmentVariable("{var_name}","{var_val}",'
            "[EnvironmentVariableTarget]::Machine)"
        )
        run_cmd_in_subprocess(["powershell.exe", pwrsh_cmd])
    elif get_os_name() == "Linux":
        run_cmd_in_subprocess(["export", f"{var_name}={var_val}"], shell=True)
    os.environ[var_name] = var_val


def remove_env_var(var_name):
    """
    remove an environment variable given its name
    """
    if get_os_name() == "Windows":
        pwrsh_cmd = (
            f'[Environment]::SetEnvironmentVariable("{var_name}",$null,'
            "[EnvironmentVariableTarget]::Machine)"
        )
        run_cmd_in_subprocess(["powershell.exe", pwrsh_cmd])
    elif get_os_name() == "Linux":
        run_cmd_in_subprocess(["unset", var_name], shell=True)
    else:
        raise NotImplementedError
    os.environ[var_name] = ""


def set_env_var_from_user_input(var_name):
    """
    set an environment variable with the given name and description based on user input
    """
    var_val = input(f"Please enter the value for environment variable {var_name}: ")
    set_env_var(var_name, var_val)


def set_env_vars(env_var_names, interactive=True):
    """
    set the necessary environment variables
    """
    variables_set = False
    for env_var_name in env_var_names:
        if (not interactive) and (
            env_var_name in ("KAFKA_PROD_CLUSTER_USERNAME", "KAFKA_PROD_CLUSTER_PASSWORD")
        ):
            continue
        if os.path.expandvars(f"${env_var_name}") == f"${env_var_name}":
            if interactive:
                set_env_var_from_user_input(env_var_name)
                variables_set = True
            else:
                raise RuntimeError(
                    f"ERROR: a value for the {env_var_name} environment variable is not set!"
                )
        else:
            if interactive:
                txt = (
                    f"A value for the {env_var_name} is already set, "
                    "would you like to reset it? [y/(n)]: "
                )
                choice = input(txt)
                if choice.lower() in ("yes", "y"):
                    set_env_var_from_user_input(env_var_name)
                    variables_set = True
    return variables_set and get_os_name() == "Windows"


def test_python_code():
    """
    briefly test the python code for the repo to catch any errors
    """
    must_rerun = set_env_vars(
        ["KAFKA_TEST_CLUSTER_USERNAME", "KAFKA_TEST_CLUSTER_PASSWORD"]
    )
    if must_rerun:
        msg = (
            "New values for environment variables have been set. "
            "Please close this window and rerun InstallService so that their values "
            "get picked up."
        )
        SERVICE_CONST.logger.info(msg)
        sys.exit(0)
    SERVICE_CONST.logger.debug("Testing code to check for errors...")
    test_script_dir_path = (
        pathlib.Path(__file__).parent.parent.parent / "test" / "test_scripts"
    )
    SERVICE_CONST.logger.debug(f"Running all tests in {test_script_dir_path}...")
    run_cmd_in_subprocess(
        [
            f"{sys.executable}",
            "-m",
            "unittest",
            "discover",
            "-s",
            f"{test_script_dir_path}",
            "-vf",
        ]
    )
    SERVICE_CONST.logger.debug("All tests complete : )")



"""
Wrapper around the KafkaCrypto producer/consumer pair communicating with the key passing topics
"""

# imports
import pathlib, warnings, logging, configparser

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from kafkacrypto import KafkaProducer, KafkaConsumer, KafkaCrypto, KafkaCryptoStore
from openmsitoolbox.utilities.misc import change_dir


class OpenMSIStreamKafkaCrypto:
    """
    A wrapper class to own and work with the Producers, Consumers, and other objects needed
    by KafkaCrypto to handle encrypting/decrypting messages and exchanging keys

    :param broker_configs: the producer/consumer-agnostic configurations for connecting
        to the broker
    :type broker_configs: dict
    :param config_file: the path to the KafkaCrypto config file for the node being used
    :type config_file: :class:`pathlib.Path`
    """

    @property
    def config_file_path(self):
        """
        The path to the KafkaCrypto config file used
        """
        return self.__config_file

    @property
    def key_serializer(self):
        """
        The crypto key serializer
        """
        return self._kc.getKeySerializer()

    @property
    def key_deserializer(self):
        """
        The crypto key deserializer
        """
        return self._kc.getKeyDeserializer()

    @property
    def value_serializer(self):
        """
        The crypto value serializer
        """
        return self._kc.getValueSerializer()

    @property
    def value_deserializer(self):
        """
        The crypto value deserializer
        """
        return self._kc.getValueDeserializer()

    def __init__(self, broker_configs, config_file):
        """
        Constructor method
        """
        kcp_cfgs, kcc_cfgs = self.__get_configs_from_file(broker_configs, config_file)
        with change_dir(pathlib.Path(config_file).parent):
            kc_logger = logging.getLogger("kafkacrypto")
            kc_logger.setLevel(logging.ERROR)
            # start up the producer and consumer
            self._kcp = KafkaProducer(**kcp_cfgs)
            self._kcc = KafkaConsumer(**kcc_cfgs)
            # initialize the KafkaCrypto object
            self._kc = KafkaCrypto(None, self._kcp, self._kcc, config_file)
            kc_logger.setLevel(logging.WARNING)
        self.__config_file = config_file

    def close(self):
        """
        Close the :class:`kafkacrypto.KafkaCrypto` object and un-set its producer/consumer
        """
        try:
            self._kc.close()
        except Exception:
            pass
        finally:
            self._kc = None
            self._kcp = None
            self._kcc = None

    def __get_configs_from_file(self, broker_configs, config_file):
        """Return the dictionaries of crypto producer and consumer configs determined
        from the KafkaCrypto config file and overwritten with the given broker configs
        from the OpenMSIStream config file. KafkaCryptoStore must be used when parsing
        the crypto config file to ensure options (and clearing of options) is properly
        handled.
        """
        # If ssl.ca.location is set in the broker configs, make sure it's written to the
        # KafkaCrypto config file as well in the right place
        if "ssl.ca.location" in broker_configs:
            config = configparser.ConfigParser(delimiters=":")
            config.read(config_file)
            section_name = f"{config_file.stem}-kafka"
            option_name = "ssl_cafile"
            if config.has_option(section_name, option_name):
                config.set(section_name, option_name, broker_configs["ssl.ca.location"])
                with open(config_file, "w") as cfg_fp:
                    config.write(cfg_fp)
        # Parse the config file and get consumer and producer configs
        cfg_parser = KafkaCryptoStore(config_file, conf_global_logger=False)
        kcc_cfgs = cfg_parser.get_kafka_config("consumer", extra="crypto")
        kcp_cfgs = cfg_parser.get_kafka_config("producer", extra="crypto")
        cfg_parser.close()
        # Overwrite with OpenMSIStream broker configs
        kcc_cfgs.update(broker_configs.copy())
        kcp_cfgs.update(broker_configs.copy())
        # KafkaCryptoStore automatically handles settings a group id if not already set
        # and makes sure unnecessary defaults (such as node_id) are not included.
        return kcp_cfgs, kcc_cfgs



"""Various utility functions used elsewhere in this subdirectory"""

# imports
import datetime
from confluent_kafka import (
    TIMESTAMP_NOT_AVAILABLE,
    TIMESTAMP_CREATE_TIME,
    TIMESTAMP_LOG_APPEND_TIME,
)
from kafkacrypto import KafkaCryptoMessage


def get_encrypted_message_timestamp_string(msg):
    """
    Given a Message object containing still-encrypted keys and/or values,
    return a string describing the timestamp in local time.
    """
    format_string = "on_%b_%d_%Y_at_%H_%M_%S_%f"
    if (
        hasattr(msg, "key")
        and hasattr(msg, "value")
        and (
            isinstance(msg.key, KafkaCryptoMessage)
            or isinstance(msg.value, KafkaCryptoMessage)
        )
    ):
        ts_type, timestamp = msg.timestamp
    else:
        ts_type, timestamp = msg.timestamp()
    if ts_type == TIMESTAMP_NOT_AVAILABLE:
        timestamp_string = f"consumed_{datetime.datetime.now().strftime(format_string)}"
    elif ts_type in (TIMESTAMP_CREATE_TIME, TIMESTAMP_LOG_APPEND_TIME):
        utc_dt = datetime.datetime.fromtimestamp(
            timestamp / 1000.0, tz=datetime.timezone.utc
        )
        local_timezone = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
        local_dt = utc_dt.astimezone(local_timezone)
        if ts_type == TIMESTAMP_CREATE_TIME:
            timestamp_string = "produced"
        elif ts_type == TIMESTAMP_LOG_APPEND_TIME:
            timestamp_string = "received"
        timestamp_string += f"_{local_dt.strftime(format_string)}"
    else:
        raise ValueError(f"ERROR: Unrecognized timestamp type {ts_type}")
    return timestamp_string


def get_encrypted_message_key_and_value_filenames(msg, topic_name):
    """
    Given a Message object containing still-encrypted keys and/or values and the name
    of the topic it came from, return the names of the files to which the key and
    value should have their bytes written
    """
    timestamp_string = get_encrypted_message_timestamp_string(msg)
    fp_pp = f"encrypted_message_from_{topic_name}_{timestamp_string}"
    return f"{fp_pp}_key.bin", f"{fp_pp}_value.bin"


def get_message_prepend(subdir_str, filename):
    """
    Return the prepend for each message key, based on the filepath
    """
    key_pp = f'{"_".join(subdir_str.split("/"))}_' if subdir_str is not None else ""
    return f"{key_pp}{filename}_chunk"



"""
Write DataFileChunks directly to disk as they are consumed from topics.
Preserve subdirectory structure if applicable
"""

# imports
import datetime, warnings

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from kafkacrypto.message import KafkaCryptoMessage
from openmsitoolbox import Runnable
from ...utilities import OpenMSIStreamArgumentParser
from ...utilities.config import RUN_CONST
from ..config import DATA_FILE_HANDLING_CONST
from ..utilities import get_encrypted_message_key_and_value_filenames
from .. import DataFileDirectory, DownloadDataFileToDisk
from .data_file_chunk_handlers import DataFileChunkProcessor


class DataFileDownloadDirectory(DataFileDirectory, DataFileChunkProcessor, Runnable):
    """
    Class representing a directory into which files are being reconstructed.

    :param dirpath: Path to the directory where reconstructed files should be saved
    :type dirpath: :class:`pathlib.Path`
    :param config_path: Path to the config file to use in defining the Broker connection
        and Consumers
    :type config_path: :class:`pathlib.Path`
    :param topic_name: Name of the topic to which the Consumers should be subscribed
    :type topic_name: str
    :param datafile_type: the type of data file that recognized files should be reconstructed as
        (must be a subclass of :class:`~.data_file_io.DownloadDataFileToDisk`)
    :type datafile_type: :class:`~.data_file_io.DownloadDataFileToDisk`, optional
    :param filepath_regex: If given, only messages associated with files whose paths match
        this regex will be consumed
    :type filepath_regex: :type filepath_regex: :func:`re.compile` or None, optional

    :raises ValueError: if `datafile_type` is not a subclass of
        :class:`~.data_file_io.DownloadDataFileToDisk`
    """

    ARGUMENT_PARSER_TYPE = OpenMSIStreamArgumentParser

    #################### PUBLIC FUNCTIONS ####################

    def __init__(
        self,
        dirpath,
        config_path,
        topic_name,
        datafile_type=DownloadDataFileToDisk,
        filepath_regex=None,
        **kwargs,
    ):
        """
        datafile_type = the type of datafile that the consumed messages are from
        In this class datafile_type should be something that extends DownloadDataFileToDisk
        """
        super().__init__(
            dirpath,
            config_path,
            topic_name,
            datafile_type=datafile_type,
            filepath_regex=filepath_regex,
            **kwargs,
        )
        if not issubclass(self.datafile_type, DownloadDataFileToDisk):
            errmsg = (
                "ERROR: DataFileDownloadDirectory requires a datafile_type that is a "
                f"subclass of DownloadDataFileToDisk but {self.datafile_type} was given!"
            )
            self.logger.error(errmsg, exc_type=ValueError)
        self.__encrypted_messages_subdir = self.dirpath / "ENCRYPTED_MESSAGES"

    def reconstruct(self):
        """
        Consumes messages and writes their data to disk using several parallel threads
        to reconstruct the files to which they correspond.
        Runs until the user inputs a command to shut it down.

        :return: the total number of messages consumed
        :rtype: int
        :return: the total number of message processed (written to disk)
        :rtype: int
        :return: the total number of completely reconstructed files
        :rtype: int
        :return: paths of up to 50 most recently reconstructed files
        :rtype: list
        """
        msg = (
            f"Will reconstruct files from messages in the {self.topic_name} topic using "
            f'{self.n_threads} thread{"s" if self.n_threads!=1 else ""}'
        )
        self.logger.info(msg)
        self.run()
        return (
            self.n_msgs_read,
            self.n_msgs_processed,
            self.n_processed_files,
            self.recent_processed_filepaths,
        )

    #################### PRIVATE HELPER FUNCTIONS ####################

    def _process_message(self, lock, msg, rootdir_to_set=None):
        retval = super()._process_message(
            lock, msg, self.dirpath if rootdir_to_set is None else rootdir_to_set
        )
        # if the message was returned because it couldn't be decrypted,
        # write it to the encrypted messages directory
        if (
            hasattr(retval, "key")
            and hasattr(retval, "value")
            and (
                isinstance(retval.key, KafkaCryptoMessage)
                or isinstance(retval.value, KafkaCryptoMessage)
            )
        ):
            with lock:
                if not self.__encrypted_messages_subdir.is_dir():
                    self.__encrypted_messages_subdir.mkdir(parents=True)
            key_fn, value_fn = get_encrypted_message_key_and_value_filenames(
                retval, self.topic_name
            )
            key_fp = self.__encrypted_messages_subdir / key_fn
            value_fp = self.__encrypted_messages_subdir / value_fn
            warnmsg = (
                "WARNING: encountered a message that failed to be decrypted. Key bytes "
                f"will be written to {key_fp.relative_to(self.dirpath)} and value bytes "
                f"will be written to {value_fp.relative_to(self.dirpath)}"
            )
            self.logger.warning(warnmsg)
            with open(key_fp, "wb") as fp:
                fp.write(bytes(retval.key))
            with open(value_fp, "wb") as fp:
                fp.write(bytes(retval.value))
            return False  # because the message wasn't processed successfully
        if retval is True:
            return retval
        # get the DataFileChunk from the message value
        try:
            dfc = msg.value()  # from a regular Kafka Consumer
        except TypeError:
            dfc = msg.value  # from KafkaCrypto
        # If the file was successfully reconstructed, return True
        if retval == DATA_FILE_HANDLING_CONST.FILE_SUCCESSFULLY_RECONSTRUCTED_CODE:
            self.logger.debug(
                f"File {dfc.relative_filepath} successfully reconstructed from stream"
            )
            with lock:
                self.recent_processed_filepaths.append(dfc.relative_filepath)
                while len(self.recent_processed_filepaths) > self.N_RECENT_FILES:
                    _ = self.recent_processed_filepaths.pop(0)
                self.n_processed_files += 1
                del self.files_in_progress_by_path[dfc.relative_filepath]
                del self.locks_by_fp[dfc.relative_filepath]
            return True
        # If the file hash was mismatched after reconstruction, return False
        if retval == DATA_FILE_HANDLING_CONST.FILE_HASH_MISMATCH_CODE:
            warnmsg = (
                f"WARNING: hashes for file {dfc.relative_filepath} not matched after "
                "reconstruction! All data have been written to disk, but not as they "
                "were uploaded."
            )
            self.logger.warning(warnmsg)
            with lock:
                del self.files_in_progress_by_path[dfc.relative_filepath]
                del self.locks_by_fp[dfc.relative_filepath]
            return False
        # if this is reached the return code was unrecognized
        self.logger.error(
            f"ERROR: unrecognized add_chunk return value ({retval})!",
            exc_type=NotImplementedError,
        )
        return False

    def _on_check(self):
        msg = (
            f"{self.n_msgs_read} messages read, {self.n_msgs_processed} messages "
            f"processed, {self.n_processed_files} files completely reconstructed so far"
        )
        self.logger.info(msg)
        if (
            len(self.files_in_progress_by_path) > 0
            or len(self.recent_processed_filepaths) > 0
        ):
            self.logger.debug(self.progress_msg)

    #################### CLASS METHODS ####################

    @classmethod
    def get_command_line_arguments(cls):
        superargs, superkwargs = super().get_command_line_arguments()
        args = [
            *superargs,
            "output_dir",
        ]
        kwargs = {**superkwargs, "n_threads": RUN_CONST.N_DEFAULT_DOWNLOAD_THREADS}
        return args, kwargs

    @classmethod
    def get_init_args_kwargs(cls, parsed_args):
        superargs, superkwargs = super().get_init_args_kwargs(parsed_args)
        args = [
            parsed_args.output_dir,
            *superargs,
        ]
        return args, superkwargs

    @classmethod
    def run_from_command_line(cls, args=None):
        """
        Run a :class:`~DataFileDownloadDirectory` directly from the command line

        Calls :func:`~reconstruct` on a :class:`~DataFileDownloadDirectory` defined by
        command line (or given) arguments

        :param args: the list of arguments to send to the parser instead of getting them
            from sys.argv
        :type args: list, optional
        """
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        # make the download directory
        init_args, init_kwargs = cls.get_init_args_kwargs(args)
        reconstructor_directory = cls(*init_args, **init_kwargs)
        # start the reconstructor running
        run_start = datetime.datetime.now()
        reconstructor_directory.logger.info(
            f"Listening for files to reconstruct in {args.output_dir}"
        )
        (
            n_read,
            n_processed,
            n_complete_files,
            complete_filepaths,
        ) = reconstructor_directory.reconstruct()
        reconstructor_directory.close()
        run_stop = datetime.datetime.now()
        # shut down when that function returns
        reconstructor_directory.logger.info(
            f"File reconstructor writing to {args.output_dir} shut down"
        )
        msg = f"{n_read} total messages were consumed"
        if len(complete_filepaths) > 0:
            msg += (
                f", {n_processed} messages were successfully processed, and "
                f'{n_complete_files} file{" was" if n_complete_files==1 else "s were"} '
                "successfully reconstructed"
            )
        else:
            msg += f" and {n_processed} messages were successfully processed"
        msg += (
            f" from {run_start} to {run_stop}\n"
            f"Most recent completed files (up to {cls.N_RECENT_FILES}):\n\t"
        )
        msg += "\n\t".join([str(filepath) for filepath in complete_filepaths])
        reconstructor_directory.logger.info(msg)


def main(args=None):
    """
    Main method to run from command line
    """
    DataFileDownloadDirectory.run_from_command_line(args)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""Reads still-encrypted message keys and values from files on disk and re-produces them
to the topic from which they came
"""

# imports
import os
from time import sleep
from argparse import ArgumentParser
from kafkacrypto import KafkaProducer, KafkaCryptoStore
from openmsitoolbox.argument_parsing.parser_callbacks import existing_file, existing_dir


def main(given_args=None):
    "Re-produce binary contents of encrypted messages to the topic from which they came"
    # Define and get command line arguments
    parser = ArgumentParser()
    parser.add_argument(
        "config", type=existing_file, help="Path to the config file to use"
    )
    parser.add_argument(
        "encrypted_messages_dir",
        type=existing_dir,
        help=(
            "Path to the directory containining encrypted message key and value files "
            "written out by a DataFileDownloadDirectory"
        ),
    )
    args = parser.parse_args(given_args)
    # Process configuration file
    kcs = KafkaCryptoStore(str(args.config), conf_global_logger=False)
    # Setup Kafka Producer *without* crypto ser (msgs are already serialized)
    kafka_config = kcs.get_kafka_config("producer")
    producer = KafkaProducer(**kafka_config)
    # Get all of the files in the encrypted files directory
    return_dir = os.getcwd()
    os.chdir(str(args.encrypted_messages_dir))
    files = sorted(filter(os.path.isfile, os.listdir(".")), key=os.path.getmtime)
    # Produce a message for each pair of encrypted key/value files
    for filename in files:
        namestart = "encrypted_message_from_"
        keyapp = "_key.bin"
        valueapp = "_value.bin"
        if filename.startswith(namestart) and filename.endswith(keyapp):
            descriptor = None
            possibilities = ("_consumed_", "_produced_", "_received_")
            for poss in possibilities:
                if poss in filename:
                    descriptor = poss
                    break
            if descriptor is None:
                raise ValueError(
                    f"ERROR: unrecognized filename format for file {filename}"
                )
            # Determine the topic from which the msg came
            # topic name cannot contain "consumed", "produced", or "received"
            tpc = filename[len(namestart) : filename.index(descriptor)]
            with open(filename, mode="rb") as ipf:  # read key
                nkv = ipf.read()
            with open(
                filename[: -len(keyapp)] + valueapp, mode="rb"
            ) as ipf:  # read value
                nvv = ipf.read()
            msg = (
                f"Reproducing {filename[:-len(keyapp)]}[{keyapp}/{valueapp}] "
                f"({len(nkv)},{len(nvv)}) to {tpc}"
            )
            print(msg)
            producer.send(tpc, key=nkv, value=nvv).get()
            producer.poll()
            sleep(0.2)
    os.chdir(return_dir)
    print("Waiting for producer to send all files to Broker...")
    producer.flush()
    producer.close()
    kcs.close()
    print("Done!")


if __name__ == "__main__":
    main()



"""Manage Linux daemons"""

# imports
import sys, os, pathlib, textwrap
from .config import SERVICE_CONST
from .utilities import run_cmd_in_subprocess
from .service_manager_base import ServiceManagerBase


class LinuxServiceManager(ServiceManagerBase):
    """
    Base class for working with Linux daemons

    :param service_name: The name of the daemon as installed
    :type service_name: str
    :param service_spec_string: A string specifying which code should be run as a daemon.
        Could be the name of an OpenMSI Runnable class, or the path to a custom Python code.
        Custom Services can also specify a :class:`OpenMSIToolbox.Runnable` class name,
        and/or a function in the file using special formatting like
        [class_name]=[path.to.file]:[function_name]. Only needed to initially install the daemon.
    :type service_spec_string: str, optional
    :param argslist: The list of arguments (as from the command line) to pass to the
        :class:`OpenMSIToolbox.Runnable` class.
        Only needed to initially install the daemon.
    :type argslist: list, optional
    :param interactive: if True, a few more messages/prompts will come up telling a user what to do
    :type interactive: bool, optional
    """

    @property
    def env_var_names(self):
        env_var_names = super().env_var_names
        # get the names of environment variables from the env_var file
        if self.env_var_filepath.is_file():
            with open(self.env_var_filepath, "r") as fp:
                lines = fp.readlines()
            for line in lines:
                linesplit = (line.strip()).split("=")
                if len(linesplit) == 2:
                    env_var_names.add(linesplit[0])
        yield from env_var_names

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.daemon_working_dir_filepath = (
            SERVICE_CONST.WORKING_DIR / f"{self.service_name}.service"
        )
        self.daemon_filepath = (
            SERVICE_CONST.DAEMON_SERVICE_DIR / self.daemon_working_dir_filepath.name
        )

    def install_service(self):
        """
        Install the daemon
        """
        super().install_service()
        # make sure systemd is running
        self.__check_systemd_installed()
        # write the daemon file pointing to the executable
        self.__write_daemon_file()
        # enable the service
        run_cmd_in_subprocess(["sudo", "systemctl", "daemon-reload"], logger=self.logger)
        run_cmd_in_subprocess(
            ["sudo", "systemctl", "enable", f"{self.service_name}.service"],
            logger=self.logger,
        )
        self.logger.info(f"Done installing {self.service_name}")

    def start_service(self):
        """
        Start the daemon
        """
        self.logger.info(f"Starting {self.service_name}...")
        self.__check_systemd_installed()
        run_cmd_in_subprocess(["sudo", "systemctl", "daemon-reload"], logger=self.logger)
        run_cmd_in_subprocess(
            ["sudo", "systemctl", "start", f"{self.service_name}.service"],
            logger=self.logger,
        )
        self.logger.info(f"Done starting {self.service_name}")

    def service_status(self):
        """
        Print the status of the daemon
        """
        self.__check_systemd_installed()
        result = run_cmd_in_subprocess(
            ["sudo", "systemctl", "status", f"{self.service_name}.service"],
            logger=self.logger,
        )
        if not result:
            warnmsg = (
                f"Failed getting status for service {self.service_name}. "
                f"Try invoking systemctl status {self.service_name}.service manually."
            )
            self.logger.warning(warnmsg)
        else:
            self.logger.info(f"{self.service_name} status: {result.decode()}")

    def stop_service(self):
        """
        Stop the daemon
        """
        self.logger.info(f"Stopping {self.service_name}...")
        self.__check_systemd_installed()
        run_cmd_in_subprocess(
            ["sudo", "systemctl", "stop", f"{self.service_name}.service"],
            logger=self.logger,
        )
        self.logger.info(f"Done stopping {self.service_name}")

    def remove_service(
        self, remove_env_vars=False, remove_install_args=False, remove_nssm=False
    ):
        """
        Remove the daemon.

        :param remove_env_vars: if True, any environment variables needed by the daemon
            will be removed.
        :type remove_env_vars: bool, optional.
        :param remove_install_args: if True, the file listing the arguments used to
            install the daemon (to make it easier to re-install) will be removed.
        :type remove_install_args: bool, optional
        :param remove_nssm: Not used for Linux daemons.
        :type remove_nssm: bool, optional
        """
        self.logger.info(f"Removing {self.service_name}...")
        self.__check_systemd_installed()
        try:
            run_cmd_in_subprocess(
                ["sudo", "systemctl", "disable", f"{self.service_name}.service"],
                logger=self.logger,
            )
        except Exception:
            pass
        if self.daemon_filepath.exists():
            run_cmd_in_subprocess(
                ["sudo", "rm", "-f", str(self.daemon_filepath)], logger=self.logger
            )
        run_cmd_in_subprocess(["sudo", "systemctl", "daemon-reload"], logger=self.logger)
        run_cmd_in_subprocess(["sudo", "systemctl", "reset-failed"], logger=self.logger)
        self.logger.info("Service removed")
        if remove_nssm:
            warnmsg = (
                "WARNING: requested to remove NSSM along with the Service, but "
                "Linux Services don't install NSSM so nothing was done."
            )
            self.logger.warning(warnmsg)
        super().remove_service(
            remove_env_vars=remove_env_vars, remove_install_args=remove_install_args
        )

    def _write_env_var_file(self):
        code = ""
        for evn in self.env_var_names:
            val = os.path.expandvars(f"${evn}")
            if val == f"${evn}":
                raise RuntimeError(
                    f"ERROR: value not found for expected environment variable {evn}!"
                )
            code += f"{evn}={val}\n"
        if code == "":
            return False
        with open(self.env_var_filepath, "w") as fp:
            fp.write(code)
        run_cmd_in_subprocess(["chmod", "go-rwx", self.env_var_filepath])
        return True

    def __check_systemd_installed(self):
        """
        Raises an error if systemd is not installed (systemd is needed to control Linux daemons)
        """
        check = run_cmd_in_subprocess(
            ["ps", "--no-headers", "-o", "comm", "1"], logger=self.logger
        )
        if check.decode().rstrip() != "systemd":
            errmsg = (
                'ERROR: Installing programs as Services ("daemons") on Linux requires systemd!'
                'You can install systemd with "sudo apt install systemd" (or similar) '
                "and try again."
            )
            self.logger.error(errmsg, exc_type=RuntimeError)

    def __write_daemon_file(self):
        """
        Write the Unit/.service file to the daemon directory that calls the Python executable
        """
        # make sure the directory to hold the file exists
        if not SERVICE_CONST.DAEMON_SERVICE_DIR.is_dir():
            SERVICE_CONST.logger.info(
                f"Creating a new daemon service directory at {SERVICE_CONST.DAEMON_SERVICE_DIR}"
            )
            SERVICE_CONST.DAEMON_SERVICE_DIR.mkdir(parents=True)
        # write out the file pointing to the python executable
        description = "Python script"
        if self.service_dict["class"] is not None:
            classdoc = self.service_dict["class"].__doc__
            if classdoc is not None:
                description = classdoc.strip().replace("\n", " ")
            else:
                description += f" for a {self.service_dict['class']} object"
        code = f"""\
            [Unit]
            Description = {description}
            Requires = network-online.target local-fs.target remote-fs.target
            After = network-online.target local-fs.target remote-fs.target

            [Service]
            Type = simple
            KillMode = none
            User = root
            ExecStart = {sys.executable} {self.exec_fp}"""
        if self.env_vars_needed:
            code += f"""\n\
            EnvironmentFile = {self.env_var_filepath}"""
        code += f"""\n\
            WorkingDirectory = {pathlib.Path().resolve()}
            Restart = always
            RestartSec = 120

            [Install]
            WantedBy = multi-user.target"""
        with open(self.daemon_working_dir_filepath, "w") as fp:
            fp.write(textwrap.dedent(code))
        run_cmd_in_subprocess(
            [
                "sudo",
                "mv",
                str(self.daemon_working_dir_filepath),
                str(self.daemon_filepath.parent),
            ],
            logger=self.logger,
        )



"""Utility class holding constants and small calculations for working with Services/daemons"""

# imports
import os, pathlib, importlib
from inspect import isclass
import pkg_resources
from openmsitoolbox.logging import OpenMSILogger


class ServicesConstants:
    """
    Constants for working with services
    """

    WORKING_DIR = (
        os.environ["OPENMSISTREAM_SERVICES_WORKING_DIR"]
        if "OPENMSISTREAM_SERVICES_WORKING_DIR" in os.environ
        else (pathlib.Path(__file__).parent / "working_dir").resolve()
    )
    NSSM_PATH = WORKING_DIR / "nssm.exe"
    # The URL to use for downloading NSSM when needed
    NSSM_DOWNLOAD_URL = "https://nssm.cc/release/nssm-2.24.zip"
    ERROR_LOG_STEM = "_ERROR_LOG.txt"
    SERVICE_EXECUTABLE_NAME_STEM = "_service_executable.py"
    DAEMON_SERVICE_DIR = pathlib.Path("/etc/systemd/system/")

    @property
    def available_services(self):
        """
        A dictionary with details of the services that are available
        """
        return self.service_dicts

    def __init__(self):
        # make the Service dictionaries to use
        self.service_dicts = []
        for script in pkg_resources.iter_entry_points("console_scripts"):
            if script.dist.key == "openmsistream":
                if script.name in (
                    "InstallService",
                    "ManageService",
                    "ProvisionNode",
                    "ReproduceUndecryptableMessages",
                    "UploadDataFile",
                ):
                    continue
                scriptstr = str(script)
                cmd = (scriptstr.split())[0]
                path = ((scriptstr.split())[2].split(":"))[0]
                funcname = (((scriptstr.split())[2]).split(":"))[1]
                module = importlib.import_module(path)
                run_classes = [
                    getattr(module, x)
                    for x in dir(module)
                    if isclass(getattr(module, x))
                    and getattr(module, x).__name__ == script.name
                ]
                if len(run_classes) != 1:
                    errmsg = (
                        f"ERROR: could not determine class for script {cmd} in file {path}! "
                        f"Possibilities found: {run_classes}"
                    )
                    raise RuntimeError(errmsg)
                self.service_dicts.append(
                    {
                        "class_name": cmd,
                        "class": run_classes[0],
                        "filepath": path,
                        "func_name": funcname,
                    }
                )
        # make the logger to use
        self.logger = OpenMSILogger(
            "Services", logger_filepath=self.WORKING_DIR / "Services.log"
        )


SERVICE_CONST = ServicesConstants()



"""
A set of Consumers sharing a subscribed topic, group ID, and (optionally)
KafkaCrypto instance for key passing
"""

# imports
import kafka
from confluent_kafka.admin import AdminClient
from confluent_kafka import TopicPartition
from openmsitoolbox import LogOwner
from .utilities import reset_to_beginning_on_assign
from .config_file_parser import KafkaConfigFileParser
from .openmsistream_consumer import OpenMSIStreamConsumer


class ConsumerGroup(LogOwner):
    """
    Class for working with a group of consumers sharing a single
    :class:`kafkacrypto.KafkaCrypto` instance

    :param config_path: Path to the config file that should be used to define Consumers in the group
    :type config_path: :class:`pathlib.Path`
    :param topic_name: The name of the topic to which the Consumers should be subscribed
    :type topic_name: str
    :param consumer_group_id: The ID string that should be used for each Consumer in the group.
        "create_new" (the default) will create a new UID to use.
    :type consumer_group_id: str, optional
    :param kafkacrypto: The :class:`~OpenMSIStreamKafkaCrypto` object that should be used
        to instantiate Consumers. Only needed if a single specific
        :class:`~OpenMSIStreamKafkaCrypto` instance should be shared.
    :type kafkacrypto: :class:`~OpenMSIStreamKafkaCrypto`, optional
    :param treat_undecryptable_as_plaintext: If True, the KafkaCrypto Deserializers
        will immediately return any keys/values that are not possibly decryptable as
        binary data. This allows faster handling of messages that will never be
        decryptable, such as when enabling or disabling encryption across a platform,
        or when unencrypted messages are mixed in a topic with encrypted messages.
    :type treat_undecryptable_as_plaintext: boolean, optional
    :param kwargs: Other keyword arguments will be added to the underlying Consumer's
        configurations, with underscores in their names replaced with dots.
    :type kwargs: dict
    """

    @property
    def topic_name(self):
        """
        Name of the topic to which Consumers are subscribed
        """
        return self.__topic_name

    @property
    def consumer_group_id(self):
        """
        String ID of Consumers in the group
        """
        return self.__consumer_group_id

    @property
    def kafkacrypto(self):
        """
        The KafkaCrypto object handling key passing and deserialization
        for the group of Consumers (if applicable)
        """
        return (
            self.__c_kwargs["kafkacrypto"] if "kafkacrypto" in self.__c_kwargs else None
        )

    def __init__(
        self,
        config_path,
        topic_name,
        *,
        consumer_group_id="create_new",
        kafkacrypto=None,
        treat_undecryptable_as_plaintext=False,
        **kwargs,
    ):
        """
        Constructor method
        """
        super().__init__(**kwargs)
        self.__group_starting_offsets = self.__get_group_starting_offsets(
            config_path, topic_name, consumer_group_id
        )
        self.__topic_name = topic_name
        self.__c_args, self.__c_kwargs = OpenMSIStreamConsumer.get_consumer_args_kwargs(
            config_path,
            group_id=consumer_group_id,
            logger=self.logger,
            kafkacrypto=kafkacrypto,
            treat_undecryptable_as_plaintext=treat_undecryptable_as_plaintext,
        )
        if len(self.__c_args) > 1 and "group.id" in self.__c_args[1]:
            self.__consumer_group_id = self.__c_args[1]["group.id"]
        else:
            self.__consumer_group_id = consumer_group_id

    def get_new_subscribed_consumer(self, *, restart_at_beginning=False, **kwargs):
        """
        Return a new Consumer, subscribed to the topic and with the shared group ID.
        Call this function from a child thread to get thread-independent Consumers.

        Note: This function just creates and subscribes the Consumer. Polling it, closing
        it, and everything else must be handled by whatever calls this function.

        :param restart_at_beginning: if True, the new Consumer will start reading partitions
            from the earliest messages available, regardless of Consumer group ID and
            auto.offset.reset values. Useful when re-reading messages.
        :type restart_at_beginning: bool, optional
        :param kwargs: other keyword arguments are passed to the
            :class:`~OpenMSIStreamConsumer` constructor method
        :type kwargs: dict

        :return: a Consumer created using the configs set in the constructor/from `kwargs`,
            subscribed to the topic
        :rtype: :class:`~OpenMSIStreamConsumer`
        """
        consumer = OpenMSIStreamConsumer(
            *self.__c_args,
            **kwargs,
            starting_offsets=self.__group_starting_offsets,
            **self.__c_kwargs,
        )
        if restart_at_beginning:
            consumer.subscribe(
                [self.__topic_name], on_assign=reset_to_beginning_on_assign
            )
        else:
            consumer.subscribe([self.__topic_name])
        return consumer

    def close(self):
        """
        Wrapper around :func:`kafkacrypto.KafkaCrypto.close`.
        """
        if "kafkacrypto" not in self.__c_kwargs:
            return
        self.__c_kwargs["kafkacrypto"].close()
        self.__c_kwargs["kafkacrypto"] = None

    def __get_group_starting_offsets(
        self, config_path, topic_name, consumer_group_id, n_retries=10
    ):
        """
        Return a list of TopicPartitions listing the starting offsets for each partition
        in the topic for the given consumer group ID

        Retries a configurable number of times if it fails at first, since there can be
        some lag involved

        Re-raises any errors encountered in getting the necessary metadata,
        returning None if that happens
        """
        caught_exc = None
        while n_retries > 0:
            cfp = KafkaConfigFileParser(config_path)
            starting_offsets = []
            try:
                cluster_metadata = AdminClient(cfp.broker_configs).list_topics(
                    topic=topic_name
                )
                n_partitions = len(cluster_metadata.topics[topic_name].partitions)
                if n_partitions <= 0:
                    raise RuntimeError(
                        f"ERROR: number of partitions for topic {topic_name} is {n_partitions}"
                    )
                kac_kwargs = {}
                for k, v in cfp.broker_configs.items():
                    if k in ("sasl.username", "sasl.password"):
                        key = k.replace(".", "_plain_")
                    else:
                        key = k.replace(".", "_")
                    kac_kwargs[key] = v
                parts = [
                    kafka.TopicPartition(topic_name, p_i) for p_i in range(n_partitions)
                ]
                tp_offsets = kafka.KafkaAdminClient(
                    **kac_kwargs
                ).list_consumer_group_offsets(
                    group_id=consumer_group_id, partitions=parts
                )
                if len(tp_offsets) != n_partitions:
                    errmsg = (
                        f"Found {n_partitions} partitions for topic {topic_name} but got "
                        f"{len(tp_offsets)} TopicPartitions listing current consumer group offsets"
                    )
                    raise RuntimeError(errmsg)
                for t_p, offset_metadata in tp_offsets.items():
                    starting_offsets.append(
                        TopicPartition(t_p.topic, t_p.partition, offset_metadata.offset)
                    )
                return starting_offsets
            except Exception as exc:
                caught_exc = exc
                n_retries -= 1
        if caught_exc:
            errmsg = (
                f'ERROR: encountered an exception when gathering initial "{topic_name}" '
                f'topic offsets for consumer group ID "{consumer_group_id}". '
                "The error will be logged below and re-raised."
            )
            self.logger.error(errmsg, exc_info=caught_exc, reraise=True)
        return None

    @classmethod
    def get_command_line_arguments(cls):
        """
        Anything extending this class should be able to access the
        "treat_undecryptable_as_plaintext" flag
        """
        superargs, superkwargs = super().get_command_line_arguments()
        args = [
            *superargs,
            "config",
            "topic_name",
            "consumer_group_id",
            "treat_undecryptable_as_plaintext",
        ]
        return args, superkwargs

    @classmethod
    def get_init_args_kwargs(cls, parsed_args):
        superargs, superkwargs = super().get_init_args_kwargs(parsed_args)
        args = [
            *superargs,
            parsed_args.config,
            parsed_args.topic_name,
        ]
        kwargs = {
            **superkwargs,
            "consumer_group_id": parsed_args.consumer_group_id,
            "treat_undecryptable_as_plaintext": parsed_args.treat_undecryptable_as_plaintext,
        }
        return args, kwargs



"""
A ConsumerAndProducerGroup whose receipt and production of messages
is managed using the ControlledProcess infrastructure
"""

# imports
import time, warnings
from abc import ABC, abstractmethod
from queue import Queue

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from kafkacrypto import KafkaCryptoMessage
    from kafkacrypto.confluent_kafka_wrapper import Message
from ..utilities.config import RUN_CONST
from ..utilities.heartbeat_producibles import MessageReproducerHeartbeatProducible
from ..utilities.controlled_processes_heartbeats import (
    ControlledProcessMultiThreadedHeartbeats,
)
from .consumer_and_producer_group import ConsumerAndProducerGroup


class ControlledMessageReproducer(
    ControlledProcessMultiThreadedHeartbeats, ConsumerAndProducerGroup, ABC
):
    """
    A ControlledProcessMultiThreaded combined with a ConsumerAndProducerGroup to
    systematically read messages from one topic and produce new messages to another.
    """

    CONSUMER_POLL_TIMEOUT = 0.050
    NO_MESSAGE_WAIT = 0.005
    # flush the producer after this many calls to produce_from_queue (could be fast)
    FLUSH_PRODUCER_EVERY = 100
    # timeout for the intermediate calls to producer.flush
    PRODUCER_FLUSH_TIMEOUT = 0.050

    def __init__(
        self,
        config_path,
        consumer_topic_name,
        producer_topic_name,
        *,
        filepath_regex=None,
        n_producer_threads=1,
        n_consumer_threads=RUN_CONST.N_DEFAULT_DOWNLOAD_THREADS,
        **kwargs,
    ):
        """
        Hang onto the number of messages read, processed, and produced
        """
        self.n_msgs_read = 0
        self.n_msgs_processed = 0
        self.n_msgs_produced = 0
        super().__init__(
            config_path,
            consumer_topic_name,
            **kwargs,
        )
        # set to true to reset new consumers to their earliest offsets
        self.restart_at_beginning = False
        # set to some regex to filter messages by their keys
        self.message_key_regex = None
        # reset the key regex after the consumer has filtered through previous messages
        self.filter_new_message_keys = False
        # set below to some regex to filter messages by filepath
        self.filepath_regex = filepath_regex
        # hold onto the last consumed message to commit its offset on shutdown
        self.last_message = None
        self.producer_topic_name = producer_topic_name
        self.n_producer_threads = n_producer_threads
        self.n_consumer_threads = n_consumer_threads
        self.producer_message_queue = Queue()
        # variables for heartbeat messages
        self.n_msgs_read_since_last_heartbeat = 0
        self.n_bytes_read_since_last_heartbeat = 0
        self.n_msgs_processed_since_last_heartbeat = 0
        self.n_bytes_processed_since_last_heartbeat = 0
        self.n_msgs_produced_since_last_heartbeat = 0
        self.n_bytes_produced_since_last_heartbeat = 0

    def get_heartbeat_message(self):
        new_msg = MessageReproducerHeartbeatProducible(
            self._heartbeat_program_id,
            self.n_msgs_read_since_last_heartbeat,
            self.n_msgs_processed_since_last_heartbeat,
            self.n_msgs_produced_since_last_heartbeat,
            self.n_bytes_read_since_last_heartbeat,
            self.n_bytes_processed_since_last_heartbeat,
            self.n_bytes_produced_since_last_heartbeat,
        )
        with self.lock:
            self.n_msgs_read_since_last_heartbeat = 0
            self.n_bytes_read_since_last_heartbeat = 0
            self.n_msgs_processed_since_last_heartbeat = 0
            self.n_bytes_processed_since_last_heartbeat = 0
            self.n_msgs_produced_since_last_heartbeat = 0
            self.n_bytes_produced_since_last_heartbeat = 0
        return new_msg

    def _run_worker(
        self,
        create_consumer=False,
        create_producer=False,
        produce_from_queue_args=None,
        produce_from_queue_kwargs=None,
    ):
        """
        Handle optional startup and shutdown of thread-independent Consumer and/or Producer.
        Serve individual messages to the _process_message function on the Consumer side,
        and produce messages from the shared Queue on the Producer side.
        """
        # create the Consumer and/or Producer for this thread
        if self.alive:
            consumer, producer = self.__get_thread_consumer_and_producer(
                create_consumer, create_producer
            )
        # start the loop for while the controlled process is alive
        calls_since_producer_flush = 0
        while self.alive:
            # if this thread has a Consumer side
            if consumer is not None:
                self.__consume_messages_while_alive(consumer)
            # if this thread has a Producer side
            if producer is not None:
                calls_since_producer_flush = self.__produce_messages_while_alive(
                    producer,
                    produce_from_queue_args,
                    produce_from_queue_kwargs,
                    calls_since_producer_flush,
                )
        # synchronously commit the offset of the last message received
        if (
            (consumer is not None)
            and (self.last_message is not None)
            and (not consumer.message_consumed_before(self.last_message))
        ):
            self.__commit_last_message_offset(consumer)
        # shut down the Consumer/Producer that was created once the process isn't alive anymore
        if consumer is not None:
            consumer.close()
        if producer is not None:
            producer.flush(timeout=-1)
            producer.close()

    def _on_shutdown(self):
        self.close()
        super()._on_shutdown()

    @abstractmethod
    def _process_message(self, lock, msg, *args, **kwargs):
        """
        Process a single message read from the thread-independent Consumer
        Returns true if processing was successful, and False otherwise

        lock = lock across all created child threads
        msg  = a single message that was consumed and should be processed by this function

        Not implemented in the base class
        """
        raise NotImplementedError

    def __get_thread_consumer_and_producer(self, create_consumer, create_producer):
        """
        Return the consumer and producer to use for a specific thread
        """
        if create_consumer:
            consumer = self.get_new_subscribed_consumer(
                restart_at_beginning=self.restart_at_beginning,
                message_key_regex=self.message_key_regex,
                filter_new_message_keys=self.filter_new_message_keys,
                filepath_regex=self.filepath_regex,
            )
        else:
            consumer = None
        if create_producer:
            producer = self.get_new_producer()
        else:
            producer = None
        if (consumer is not None) and (
            ("enable.auto.commit" not in consumer.configs.keys())
            or (consumer.configs["enable.auto.commit"] is True)
        ):
            warnmsg = (
                "WARNING: enable.auto.commit has not been set to False for a Consumer "
                "that will manually commit offsets. Missed or duplicate messages could result. "
                'You can set "enable.auto.commit"=False in the "consumer" section of '
                "the config file to re-enable manual offset commits (recommended)."
            )
            self.logger.warning(warnmsg)
        return consumer, producer

    def __consume_messages_while_alive(self, consumer):
        """
        Runs continuously while the calling thread is alive to consume messages
        """
        # consume a message from the topic
        msg = consumer.get_next_message(self.CONSUMER_POLL_TIMEOUT)
        if msg is None:
            time.sleep(self.NO_MESSAGE_WAIT)  # wait just a bit to not over-tax things
            return
        with self.lock:
            self.n_msgs_read += 1
            self.n_msgs_read_since_last_heartbeat += 1
            if (
                hasattr(msg, "key")
                and hasattr(msg, "value")
                and (
                    isinstance(msg.key, KafkaCryptoMessage)
                    or isinstance(msg.value, KafkaCryptoMessage)
                )
            ):
                self.n_bytes_read_since_last_heartbeat += len(bytes(msg))
            elif isinstance(msg, Message):
                self.n_bytes_read_since_last_heartbeat += len(msg.value)
            else:
                self.n_bytes_read_since_last_heartbeat += len(msg)
            self.last_message = msg
        # send the message to the _process_message function
        retval = self._process_message(self.lock, msg)
        # count and (asynchronously) commit the message as processed
        if retval:
            with self.lock:
                self.n_msgs_processed += 1
                self.n_msgs_processed_since_last_heartbeat += 1
                if (
                    hasattr(msg, "key")
                    and hasattr(msg, "value")
                    and (
                        isinstance(msg.key, KafkaCryptoMessage)
                        or isinstance(msg.value, KafkaCryptoMessage)
                    )
                ):
                    self.n_bytes_processed_since_last_heartbeat += len(bytes(msg))
                elif isinstance(msg, Message):
                    self.n_bytes_processed_since_last_heartbeat += len(msg.value)
                else:
                    self.n_bytes_processed_since_last_heartbeat += len(msg)
            if not consumer.message_consumed_before(msg):
                tps = consumer.commit(msg)
                if tps is None:
                    return
                for t_p in tps:
                    if t_p.error is not None:
                        warnmsg = (
                            f"WARNING: failed to synchronously commit offset of last message "
                            f'received on "{t_p.topic}" partition {t_p.partition}. '
                            "Duplicate messages may result when this Consumer is restarted. "
                            f"Error reason: {t_p.error.str()}"
                        )
                        self.logger.warning(warnmsg)

    def __produce_messages_while_alive(
        self, producer, pfq_args, pfq_kwargs, calls_since_flush
    ):
        """
        Runs continuously while the calling thread is alive to produce messages
        """
        if self.producer_message_queue.empty():
            time.sleep(self.NO_MESSAGE_WAIT)  # wait just a bit to not overtax things
        else:
            pfq_args = [] if pfq_args is None else pfq_args
            pfq_kwargs = {} if pfq_kwargs is None else pfq_kwargs
            producer.produce_from_queue(
                self.producer_message_queue,
                self.producer_topic_name,
                *pfq_args,
                **pfq_kwargs,
            )
        if calls_since_flush >= self.FLUSH_PRODUCER_EVERY:
            producer.flush(timeout=self.PRODUCER_FLUSH_TIMEOUT)
            calls_since_flush = 0
        else:
            calls_since_flush += 1
        return calls_since_flush

    def __commit_last_message_offset(self, consumer):
        """
        Commit the offset of the last message received while blocking
        """
        try:
            tps = consumer.commit(self.last_message, asynchronous=False)
            if tps is None:
                return
            for t_p in tps:
                if t_p.error is not None:
                    warnmsg = (
                        f"WARNING: failed to synchronously commit offset of last message "
                        f'received on "{t_p.topic}" partition {t_p.partition}. '
                        "Duplicate messages may result when this Consumer is restarted. "
                        f"Error reason: {t_p.error.str()}"
                    )
                    self.logger.warning(warnmsg)
        except Exception as exc:
            warnmsg = (
                "WARNING: failed to synchronously commit offset of last message received. "
                "Duplicate messages may be read the next time this Consumer is started. "
                "Error will be logged below but not re-raised."
            )
            self.logger.warning(warnmsg, exc_info=exc)

    @classmethod
    def get_command_line_arguments(cls):
        superargs, superkwargs = super().get_command_line_arguments()
        try:
            superkwargs.pop("n_threads")
        except KeyError:
            pass
        try:
            superargs.pop(superargs.index("n_threads"))
        except ValueError:
            pass
        args = [
            *superargs,
            "producer_topic_name",
            "n_consumer_threads",
            "n_producer_threads",
            "download_regex",
        ]
        return args, superkwargs

    @classmethod
    def get_init_args_kwargs(cls, parsed_args):
        parsed_args.n_threads = max(
            parsed_args.n_producer_threads, parsed_args.n_consumer_threads
        )
        superargs, superkwargs = super().get_init_args_kwargs(parsed_args)
        args = [
            *superargs,
            parsed_args.producer_topic_name,
        ]
        kwargs = {
            **superkwargs,
            "n_producer_threads": parsed_args.n_producer_threads,
            "n_consumer_threads": parsed_args.n_consumer_threads,
            "filepath_regex": parsed_args.download_regex,
        }
        return args, kwargs



"""
A wrapper around KafkaCrypto provision scripts to automatically
move the output in the location expected by OpenMSIStream
"""

# imports
import pathlib, shutil, logging, warnings
from argparse import ArgumentParser
import urllib.request

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import kafkacrypto
from openmsitoolbox.logging import OpenMSILogger
from openmsitoolbox.utilities.misc import change_dir
from ..utilities.config import RUN_CONST
from ..utilities.config_file_parser import ConfigFileParser

# constants
LOGGER = OpenMSILogger("ProvisionNode", logging.INFO)
KC_PATH = kafkacrypto.__path__
SP_NAME = "simple-provision.py"
OP_NAME = "online-provision.py"
GITHUB_PREPEND = "https://raw.githubusercontent.com/"
GITHUB_URL = f"{GITHUB_PREPEND}tmcqueen-materials/kafkacrypto/master/tools/{SP_NAME}"
TEMP_DIR_PATH = RUN_CONST.CONFIG_FILE_DIR / "temp_kafkacrypto_dir"


def get_script_location_and_code_from_input(args):
    """
    Use a given path to get the script code
    """
    p_code = None
    p_loc = None
    if args.script_path.is_file():
        p_loc = args.script_path
        with open(args.script_path, "rb") as fp:
            p_code = fp.read()
    # directory containing the script was given
    elif args.script_path.is_dir():
        if (args.script_path / SP_NAME).is_file():
            p_loc = args.script_path / SP_NAME
            with open(args.script_path / SP_NAME, "rb") as fp:
                p_code = fp.read()
        elif (args.script_path / OP_NAME).is_file():
            p_loc = args.script_path / OP_NAME
            with open(args.script_path / OP_NAME, "rb") as fp:
                p_code = fp.read()
    return p_loc, p_code


def get_script_location_and_code(args):
    """
    Return the location of the provision script and its text contents, given the arguments
    """
    p_code = None
    p_loc = None
    if args.script_path is not None:
        # path to the script itself was given
        p_loc, p_code = get_script_location_and_code_from_input(args)
    # if not set yet, try getting from the kafkacrypto install location
    if len(KC_PATH) == 1:
        if args.mode == "simple":
            to_try = pathlib.Path(KC_PATH[0]).parent / "tools" / SP_NAME
        elif args.mode == "online":
            to_try = pathlib.Path(KC_PATH[0]).parent / "tools" / OP_NAME
        if to_try.is_file():
            p_loc = to_try
            with open(to_try, "rb") as fp:
                p_code = fp.read()
    if p_code is None:
        # if all else fails, try getting from the Github webpage
        if args.mode == "simple":
            try:
                with urllib.request.urlopen(GITHUB_URL) as urlp:
                    p_code = urlp.read()
                p_loc = GITHUB_URL
            except Exception:
                pass
    if p_loc is None or p_code is None:
        LOGGER.error(
            "ERROR: failed to find the provisioning script to use!", exc_type=RuntimeError
        )
    return p_loc, p_code


def move_files(p_loc):
    """
    Move the created files into the expected locations
    """
    try:
        new_files = {}
        exts = [".config", ".seed", ".crypto"]
        for ext in exts:
            n_files = 0
            for fp in TEMP_DIR_PATH.glob(f"*{ext}"):
                n_files += 1
                new_files[ext] = fp.resolve()
            if n_files != 1:
                LOGGER.error(
                    f"ERROR: found {n_files} new {ext} files in {TEMP_DIR_PATH}",
                    exc_type=RuntimeError,
                )
        node_id = None
        for ext in exts:
            filename = new_files[ext].name
            this_node_id = ((filename).split(ext))[0]
            if node_id is None:
                node_id = this_node_id
            elif node_id != this_node_id:
                errmsg = (
                    f"ERROR: found a file called {filename} "
                    f"that conflicts with node_id {node_id}!"
                )
                LOGGER.error(errmsg, exc_type=RuntimeError)
        cfp = ConfigFileParser(new_files[".config"], logger=LOGGER)
        default_dict = cfp.get_config_dict_for_groups("DEFAULT")
        if "node_id" not in default_dict:
            LOGGER.error(
                f"ERROR: node_id not listed in {new_files['.config']}!",
                exc_type=ValueError,
            )
        elif default_dict["node_id"] != node_id:
            errmsg = (
                f"ERROR: node_id listed in {new_files['.config']} mismatched "
                f"to filenames ({node_id})!"
            )
            LOGGER.error(errmsg, exc_type=ValueError)
        new_dirpath = TEMP_DIR_PATH.parent / node_id
        TEMP_DIR_PATH.rename(new_dirpath)
        LOGGER.info(f'Successfuly set up new KafkaCrypto node called "{node_id}"')
    except Exception as exc:
        errmsg = (
            f"ERROR: Running {p_loc} did not produce the expected output! "
            "Temporary directories will be removed and you will need to try again. "
            "Exception will be logged and re-raised."
        )
        LOGGER.error(errmsg, exc_info=exc, reraise=True)
    finally:
        if TEMP_DIR_PATH.is_dir():
            shutil.rmtree(TEMP_DIR_PATH)


def main():
    """
    Script to run either simple or online provision for KafkaCrypto and
    move the output to the location expected by OpenMSIStream
    """
    # command line arguments
    parser = ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--mode",
        choices=["simple", "online"],
        default="simple",
        help="""Choice of which known type of provisioning to use.
                               Can also provide a path to the script to run instead.""",
    )
    group.add_argument(
        "--script-path",
        type=pathlib.Path,
        default=".",
        help="Path to the provision script to run",
    )
    args = parser.parse_args()
    if TEMP_DIR_PATH.is_dir():
        shutil.rmtree(TEMP_DIR_PATH)
    # get the location/content of the simple-provision script
    p_loc, p_code = get_script_location_and_code(args)
    # run the script
    try:
        if not TEMP_DIR_PATH.is_dir():
            TEMP_DIR_PATH.mkdir(parents=True)
        with change_dir(TEMP_DIR_PATH):
            exec(p_code)
    except Exception as exc:
        errmsg = (
            f"ERROR: failed to run provisioning using {p_loc}! "
            "Exception will be logged and re-raised."
        )
        LOGGER.error(errmsg, exc_info=exc, reraise=True)
    # make sure required files exist and move them into a new directory named for the node_ID
    move_files(p_loc)


if __name__ == "__main__":
    main()



" ControlledProcess classes that periodically also produce messages to heartbeat topics "

# imports
import datetime
from abc import ABC
from confluent_kafka.serialization import StringSerializer
from confluent_kafka import SerializingProducer
from openmsitoolbox.argument_parsing.has_arguments import HasArguments
from openmsitoolbox.controlled_process.controlled_process import ControlledProcess
from openmsitoolbox import ControlledProcessSingleThread, ControlledProcessMultiThreaded
from ..kafka_wrapper.config_file_parser import KafkaConfigFileParser
from ..kafka_wrapper import OpenMSIStreamProducer
from .heartbeat_producibles import HeartbeatProducible


class ControlledProcessHeartbeats(ControlledProcess, HasArguments, ABC):
    "A long-running process that occasionally produces messages to a heartbeat topic"

    def __init__(
        self,
        config_path,
        *args,
        heartbeat_topic_name=None,
        heartbeat_program_id=None,
        heartbeat_interval_secs=None,
        **kwargs,
    ):
        try:
            super().__init__(config_path, *args, **kwargs)
        except TypeError:
            super().__init__(*args, **kwargs)
        self.__heartbeat_topic_name = heartbeat_topic_name
        self._heartbeat_program_id = heartbeat_program_id
        self.__heartbeat_interval_secs = heartbeat_interval_secs
        self.__heartbeat_producer = None
        self.__last_heartbeat = datetime.datetime.now()
        # just return if heartbeats won't be produced
        if self.__heartbeat_topic_name is None:
            return
        cfp = KafkaConfigFileParser(config_path, logger=self.logger)
        if "heartbeat" in cfp.available_group_names:
            heartbeat_config_dict = cfp.heartbeat_configs
        else:
            self.logger.warning(
                (
                    f"WARNING: config file at {config_path} has no 'heartbeat' section but "
                    "a heartbeat topic name was given."
                ),
            )
            heartbeat_config_dict = {}
        if "key.serializer" not in heartbeat_config_dict:
            heartbeat_config_dict["key.serializer"] = StringSerializer()
        if "value.serializer" not in heartbeat_config_dict:
            heartbeat_config_dict["value.serializer"] = StringSerializer()
        all_heartbeat_producer_configs = {}
        if "bootstrap.servers" not in heartbeat_config_dict:
            all_heartbeat_producer_configs.update(cfp.broker_configs)
        all_heartbeat_producer_configs.update(heartbeat_config_dict)
        self.__heartbeat_producer = OpenMSIStreamProducer(
            SerializingProducer, all_heartbeat_producer_configs, logger=self.logger
        )
        self.logger.info(
            (
                "Heartbeat messages will be produced to the "
                f"'{self.__heartbeat_topic_name}' topic every "
                f"{self.__heartbeat_interval_secs} seconds"
            )
        )
        if self._heartbeat_program_id is None:
            self._heartbeat_program_id = self.__heartbeat_producer.producer_id

    def get_heartbeat_message(self):
        """Return the HeartbeatProducible-type object that should be produced to the
        heartbeat topic
        """
        return HeartbeatProducible(self._heartbeat_program_id)

    def _print_still_alive(self):
        """Print the "still alive" character to the console like a regular
        ControlledProcess, but also produce messages to the heartbeat topic if applicable
        """
        super()._print_still_alive()
        if (
            self.__heartbeat_producer is not None
            and (datetime.datetime.now() - self.__last_heartbeat).total_seconds()
            > self.__heartbeat_interval_secs
        ):
            self.__heartbeat_producer.produce_object(
                self.get_heartbeat_message(), self.__heartbeat_topic_name
            )
            self.__last_heartbeat = datetime.datetime.now()

    def _on_shutdown(self):
        try:
            super()._on_shutdown()
        except NotImplementedError:
            pass
        if self.__heartbeat_producer is not None:
            # send a final heartbeat
            self.__heartbeat_producer.produce_object(
                self.get_heartbeat_message(), self.__heartbeat_topic_name
            )
            self.logger.info("Flushing heartbeat producer")
            self.__heartbeat_producer.flush()
            self.__heartbeat_producer.close()

    @classmethod
    def get_command_line_arguments(cls):
        superargs, superkwargs = super().get_command_line_arguments()
        args = [
            *superargs,
            "heartbeat_topic_name",
            "heartbeat_program_id",
            "heartbeat_interval_secs",
        ]
        return args, superkwargs

    @classmethod
    def get_init_args_kwargs(cls, parsed_args):
        superargs, superkwargs = super().get_init_args_kwargs(parsed_args)
        kwargs = {
            **superkwargs,
            "heartbeat_topic_name": parsed_args.heartbeat_topic_name,
            "heartbeat_program_id": parsed_args.heartbeat_program_id,
            "heartbeat_interval_secs": parsed_args.heartbeat_interval_secs,
        }
        return superargs, kwargs


class ControlledProcessSingleThreadHeartbeats(
    ControlledProcessHeartbeats, ControlledProcessSingleThread
):
    """A long-running process with a single worker thread that occasionally produces
    messages to a heartbeat topic
    """


class ControlledProcessMultiThreadedHeartbeats(
    ControlledProcessHeartbeats, ControlledProcessMultiThreaded
):
    """A long-running process with multiple worker threads that occasionally produces
    messages to a heartbeat topic
    """
